# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from http import server as SimpleHTTPServer
import json
import multiprocessing
import os
import shutil
import socketserver
import tempfile
import time

from oslo_utils import uuidutils
from tempest.lib.cli import output_parser
from tempest.lib import exceptions

from muranoclient.tests.functional.cli import utils
from muranoclient.tests.functional import muranoclient


class CLIUtilsTestBase(muranoclient.ClientTestBase):
    """Basic methods for Murano CLI client."""

    def delete_murano_object(self, murano_object, obj_to_del):
        """Delete Murano object

        Delete Murano object like environment, category or
        environment-template.
        """
        if obj_to_del not in self.listing('{0}-list'.format(murano_object)):
            return
        object_list = self.listing('{0}-delete'.format(murano_object),
                                   params=obj_to_del['ID'])
        start_time = time.time()
        while obj_to_del in self.listing('{0}-list'.format(murano_object)):
            if start_time - time.time() > 60:
                self.fail("{0} is not deleted in 60 seconds".
                          format(murano_object))
        return object_list

    def create_murano_object(self, murano_object, prefix_object_name):
        """Create Murano object

        Create Murano object like environment, category or
        environment-template.
        """
        object_name = self.generate_name(prefix_object_name)
        mrn_objects = self.listing('{0}-create'.format(murano_object),
                                   params=object_name)
        mrn_object = None
        for obj in mrn_objects:
            if object_name == obj['Name']:
                mrn_object = obj
                break
        if mrn_object is None:
            self.fail("Murano {0} has not been created!".format(murano_object))

        self.addCleanup(self.delete_murano_object, murano_object, mrn_object)
        return mrn_object

    def create_murano_object_parameter(self, murano_object, prefix_object_name,
                                       param):
        """Create Murano object

        Create Murano object like environment, category or
        environment-template.
        """
        object_name = self.generate_name(prefix_object_name)
        params = '{0} {1}'.format(param, object_name)

        mrn_objects = self.listing('{0}-create'.format(murano_object),
                                   params=params)
        mrn_object = None
        for obj in mrn_objects:
            if object_name == obj['Name']:
                mrn_object = obj
                break
        if mrn_object is None:
            self.fail("Murano {0} has not been created!".format(murano_object))

        self.addCleanup(self.delete_murano_object, murano_object, mrn_object)
        return mrn_object

    @staticmethod
    def generate_uuid():
        """Generate uuid for objects."""
        return uuidutils.generate_uuid(dashed=False)

    @staticmethod
    def generate_name(prefix):
        """Generate name for objects."""
        suffix = CLIUtilsTestBase.generate_uuid()[:8]
        return "{0}_{1}".format(prefix, suffix)

    def get_table_struct(self, command, params=""):
        """Get table structure i.e. header of table."""
        return output_parser.table(self.murano(command,
                                               params=params))['headers']

    def get_object(self, object_list, object_value):
        """"Get Murano object by value from list of Murano objects."""
        for obj in object_list:
            if object_value in obj.values():
                return obj

    def get_property_value(self, obj, prop):
        return [o['Value'] for o in obj
                if o['Property'] == '{0}'.format(prop)][0]


class TestSuiteRepository(CLIUtilsTestBase):

    def setUp(self):
        super(TestSuiteRepository, self).setUp()
        self.serve_dir = tempfile.mkdtemp(suffix="repo")
        self.app_name = self.generate_name("dummy_app")
        self.dummy_app_path = self._compose_app(name=self.app_name)

    def tearDown(self):
        super(TestSuiteRepository, self).tearDown()
        shutil.rmtree(self.serve_dir)

    def run_server(self):
        def serve_function():
            class Handler(SimpleHTTPServer.SimpleHTTPRequestHandler):
                pass
            os.chdir(self.serve_dir)
            httpd = socketserver.TCPServer(
                ("0.0.0.0", 8089),
                Handler, bind_and_activate=False)
            httpd.allow_reuse_address = True
            httpd.server_bind()
            httpd.server_activate()
            httpd.serve_forever()
        self.p = multiprocessing.Process(target=serve_function)
        self.p.start()

    def stop_server(self):
        self.p.terminate()

    def _compose_app(self, name, require=None):
        package_dir = os.path.join(self.serve_dir, 'apps/', name)
        shutil.copytree(os.path.join(os.path.dirname(
            os.path.realpath(__file__)), 'MockApp'), package_dir)

        app_name = utils.compose_package(
            name,
            os.path.join(package_dir, 'manifest.yaml'),
            package_dir,
            require=require,
            archive_dir=os.path.join(self.serve_dir, 'apps/'),
        )

        return app_name


class CLIUtilsTestPackagesBase(TestSuiteRepository):
    """Basic methods for Murano Packages CLI client."""

    def import_package(self, pkg_name, pkg_path, *args):
        """Create Murano dummy package and import it by url."""

        actions = ' '.join(args)
        params = '{0} {1}'.format(pkg_path, actions)
        package = self.listing('package-import', params=params)
        package = self.get_object(package, pkg_name)
        self.addCleanup(self.delete_murano_object, 'package', package)

        return package

    def prepare_file_with_obj_model(self, obj_model):
        temp_file = tempfile.NamedTemporaryFile(prefix="murano-obj-model",
                                                delete=False)
        self.addCleanup(os.remove, temp_file.name)

        with open(temp_file.name, 'w') as tf:
            tf.write(json.dumps([obj_model]))

        return temp_file.name

    def wait_deployment_result(self, env_id, timeout=180):
        start_time = time.time()

        env = self.listing('environment-show', params=env_id)
        env_status = self.get_property_value(env, 'status')

        expected_statuses = ['ready', 'deploying']

        while env_status != 'ready':
            if time.time() - start_time > timeout:
                msg = ("Environment exceeds timeout {0} to change state "
                       "to Ready. Environment: {1}".format(timeout, env))
                raise exceptions.TimeoutException(msg)

            env = self.listing('environment-show', params=env_id)
            env_status = self.get_property_value(env, 'status')

            if env_status not in expected_statuses:
                msg = ("Environment status %s is not in expected "
                       "statuses: %s" % (env_status, expected_statuses))
                raise exceptions.TempestException(msg)

            time.sleep(2)

        return True

    def prepare_bundle_with_non_existed_package(self):
        temp_file = tempfile.NamedTemporaryFile(mode='w',
                                                delete=False)
        self.addCleanup(os.remove, temp_file.name)

        with open(temp_file.name, 'w') as tf:
            tf.write(json.dumps({'Packages': [
                {'Name': 'first_app'},
                {'Name': 'second_app', 'Version': '1.0'}
            ]}))

        return temp_file.name

    def prepare_bundle_with_invalid_format(self):
        temp_file = tempfile.NamedTemporaryFile(mode='w', delete=False)
        self.addCleanup(os.remove, temp_file.name)

        with open(temp_file.name, 'w') as tf:
            tf.write('Packages: [{Name: first_app}, {Name: second_app}]')

        return temp_file.name

    def deploy_environment(self, env_id, obj_model):
        session = self.listing('environment-session-create',
                               params=env_id)
        session_id = self.get_property_value(session, 'id')

        temp_file = self.prepare_file_with_obj_model(obj_model)

        self.listing('environment-apps-edit',
                     params='--session-id {0} {1} {2}'.
                     format(session_id, env_id, temp_file))

        self.listing('environment-deploy',
                     params='{0} --session-id {1}'.
                     format(env_id, session_id))

        result = self.wait_deployment_result(env_id)
        self.assertTrue(result)



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import fixtures
import testtools


class TestCaseShell(testtools.TestCase):
    TEST_REQUEST_BASE = {
        'verify': True,
    }

    def setUp(self):
        super(TestCaseShell, self).setUp()
        if (os.environ.get('OS_STDOUT_CAPTURE') == 'True' or
                os.environ.get('OS_STDOUT_CAPTURE') == '1'):
            stdout = self.useFixture(fixtures.StringStream('stdout')).stream
            self.useFixture(fixtures.MonkeyPatch('sys.stdout', stdout))
        if (os.environ.get('OS_STDERR_CAPTURE') == 'True' or
                os.environ.get('OS_STDERR_CAPTURE') == '1'):
            stderr = self.useFixture(fixtures.StringStream('stderr')).stream
            self.useFixture(fixtures.MonkeyPatch('sys.stderr', stderr))


class TestAdditionalAsserts(testtools.TestCase):

    def check_dict_is_subset(self, dict1, dict2):
        # There is an assert for this in Python 2.7 but not 2.6
        self.assertTrue(all(k in dict2 and dict2[k] == v for k, v
                            in dict1.items()))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Application-catalog v1 deployments action implementation"""

from osc_lib.command import command
from osc_lib import utils
from oslo_log import log as logging

from muranoclient.apiclient import exceptions

LOG = logging.getLogger(__name__)


class ListDeployment(command.Lister):
    """List deployments for an environment."""

    def get_parser(self, prog_name):
        parser = super(ListDeployment, self).get_parser(prog_name)
        parser.add_argument(
            "id",
            metavar="<ID>",
            nargs="?",
            default=None,
            help=("Environment ID for which to list deployments."),
        )
        parser.add_argument(
            "--all-environments",
            action="store_true",
            default=False,
            help="List all deployments for all environments in user's tenant."
        )

        return parser

    def take_action(self, parsed_args):
        LOG.debug("take_action({0})".format(parsed_args))
        client = self.app.client_manager.application_catalog

        all_environments = getattr(parsed_args, 'all_environments', False)
        env_id = getattr(parsed_args, 'id', None)

        if env_id and all_environments:
            raise exceptions.CommandError(
                'Environment ID and all-environments flag cannot both be set.')
        elif not env_id and not all_environments:
            raise exceptions.CommandError(
                'Either environment ID or all-environments flag must be set.')

        if all_environments:
            data = client.deployments.list(None, all_environments)
        else:
            environment = utils.find_resource(client.environments,
                                              env_id)
            data = client.deployments.list(environment.id)

        columns = ('id', 'state', 'created', 'updated', 'finished')
        column_headers = [c.capitalize() for c in columns]
        return (
            column_headers,
            list(utils.get_item_properties(
                s,
                columns,
            ) for s in data)
        )



# Copyright (c) 2016 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class StaticActionResult(object):
    def __init__(self, result, exception=None):
        self._result = result
        self._exception = exception

    def get_result(self):
        if self._exception:
            raise self._exception
        return self._result

    def check_result(self):
        return True


# Not a true manager yet; should be changed to be one if CRUD
# functionality becomes available for actions.
class StaticActionManager(object):
    def __init__(self, api):
        self.api = api

    def call(self, arguments):
        url = '/v1/actions'
        try:
            resp, body = self.api.json_request(url, 'POST', data=arguments)
            return StaticActionResult(body)
        except Exception as e:
            if e.code >= 500:
                raise
            return StaticActionResult(None, exception=e)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import io
import json
import os
import shutil
import sys
import tempfile
from unittest import mock

from testtools import matchers

from muranoclient.common import exceptions as common_exceptions
from muranoclient.common import utils as mc_utils
from muranoclient.osc.v1 import package as osc_pkg
from muranoclient.tests.unit.osc.v1 import fakes
from muranoclient.tests.unit import test_utils
from muranoclient.v1 import packages

from osc_lib import exceptions as exc
from osc_lib import utils
import requests_mock

make_pkg = test_utils.make_pkg

FIXTURE_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                           'fixture_data'))

COLUMNS = ['Id', 'Name', 'Fully_qualified_name', 'Author', 'Active',
           'Is public', 'Type', 'Version']

DATA = {
    'class_definitions': ['com.example.apache.ApacheHttpServer'],
    'updated': '2016-09-20T06:23:45.000000',
    'description': 'Test description.\n',
    'created': '2016-09-20T06:23:15.000000',
    'author': 'Mirantis, Inc',
    'enabled': True,
    'owner_id': 'a203405ea871484a940850d6c0b8dfd9',
    'tags': ['Server', 'WebServer', 'Apache', 'HTTP', 'HTML'],
    'is_public': False,
    'fully_qualified_name': 'com.example.apache.ApacheHttpServer',
    'type': 'Application',
    'id': '46860070-5f8a-4936-96e8-d7b89e5187d7',
    'categories': [],
    'name': 'Apache HTTP Server'
}


class TestPackage(fakes.TestApplicationCatalog):
    def setUp(self):
        super(TestPackage, self).setUp()
        self.package_mock = self.app.client_manager.application_catalog.\
            packages
        self.package_mock.reset_mock()


class TestCreatePackage(TestPackage):
    def setUp(self):
        super(TestCreatePackage, self).setUp()

        # Command to test
        self.cmd = osc_pkg.CreatePackage(self.app, None)

    def test_create_package_without_args(self):
        arglist = []
        parsed_args = self.check_parser(self.cmd, arglist, [])
        error = self.assertRaises(exc.CommandError,
                                  self.cmd.take_action, parsed_args)
        self.assertEqual('Provide --template for a HOT-based package, OR at '
                         'least --classes-dir for a MuranoPL-based package',
                         str(error))

    def test_create_package_template_and_classes_args(self):
        heat_template = os.path.join(FIXTURE_DIR, 'heat-template.yaml')
        classes_dir = os.path.join(FIXTURE_DIR, 'test-app', 'Classes')
        arglist = ['--template', heat_template, '--classes-dir', classes_dir]
        parsed_args = self.check_parser(self.cmd, arglist, [])
        error = self.assertRaises(exc.CommandError,
                                  self.cmd.take_action, parsed_args)
        self.assertEqual('Provide --template for a HOT-based package, OR'
                         ' --classes-dir for a MuranoPL-based package',
                         str(error))

    def test_create_hot_based_package(self):
        with tempfile.NamedTemporaryFile() as f:
            RESULT_PACKAGE = f.name
            heat_template = os.path.join(FIXTURE_DIR, 'heat-template.yaml')
            logo = os.path.join(FIXTURE_DIR, 'logo.png')
            arglist = ['--template', heat_template, '--output', RESULT_PACKAGE,
                       '-l', logo]
            parsed_args = self.check_parser(self.cmd, arglist, [])
            orig = sys.stdout
            try:
                sys.stdout = io.StringIO()
                self.cmd.take_action(parsed_args)
            finally:
                stdout = sys.stdout.getvalue()
                sys.stdout.close()
                sys.stdout = orig
            matchers.MatchesRegex(stdout,
                                  "Application package "
                                  "is available at {0}".format(RESULT_PACKAGE))

    def test_create_mpl_package(self):
        with tempfile.NamedTemporaryFile() as f:
            RESULT_PACKAGE = f.name
            classes_dir = os.path.join(FIXTURE_DIR, 'test-app', 'Classes')
            resources_dir = os.path.join(FIXTURE_DIR, 'test-app', 'Resources')
            ui = os.path.join(FIXTURE_DIR, 'test-app', 'ui.yaml')
            arglist = ['-c', classes_dir, '-r', resources_dir,
                       '-u', ui, '-o', RESULT_PACKAGE]
            parsed_args = self.check_parser(self.cmd, arglist, [])
            orig = sys.stdout
            try:
                sys.stdout = io.StringIO()
                self.cmd.take_action(parsed_args)
            finally:
                stdout = sys.stdout.getvalue()
                sys.stdout.close()
                sys.stdout = orig
            matchers.MatchesRegex(stdout,
                                  "Application package "
                                  "is available at {0}".format(RESULT_PACKAGE))


class TestPackageList(TestPackage):

    def setUp(self):
        super(TestPackageList, self).setUp()
        self.cmd = osc_pkg.ListPackages(self.app, None)
        self.package_mock.filter.return_value = \
            [packages.Package(None, DATA)]
        utils.get_dict_properties = mock.MagicMock(return_value='')

    def test_package_list_defaults(self):
        arglist = []
        parsed_args = self.check_parser(self.cmd, arglist, [])

        columns, data = self.cmd.take_action(parsed_args)

        self.package_mock.filter.assert_called_with(
            include_disabled=False,
            owned=False)
        self.assertEqual(COLUMNS, columns)

    def test_package_list_with_limit(self):
        arglist = ['--limit', '10']
        parsed_args = self.check_parser(self.cmd, arglist, [])

        columns, data = self.cmd.take_action(parsed_args)

        self.package_mock.filter.assert_called_with(
            include_disabled=False,
            limit=10,
            owned=False)

    def test_package_list_with_marker(self):
        arglist = ['--marker', '12345']
        parsed_args = self.check_parser(self.cmd, arglist, [])

        columns, data = self.cmd.take_action(parsed_args)

        self.package_mock.filter.assert_called_with(
            include_disabled=False,
            marker='12345',
            owned=False)

    def test_package_list_with_name(self):
        arglist = ['--name', 'mysql']
        parsed_args = self.check_parser(self.cmd, arglist, [])

        columns, data = self.cmd.take_action(parsed_args)

        self.package_mock.filter.assert_called_with(
            include_disabled=False,
            name='mysql',
            owned=False)

    def test_package_list_with_fqn(self):
        arglist = ['--fqn', 'mysql']
        parsed_args = self.check_parser(self.cmd, arglist, [])

        columns, data = self.cmd.take_action(parsed_args)

        self.package_mock.filter.assert_called_with(
            include_disabled=False,
            fqn='mysql',
            owned=False)


class TestPackageDelete(TestPackage):
    def setUp(self):
        super(TestPackageDelete, self).setUp()
        self.package_mock.delete.return_value = None
        self.package_mock.filter.return_value = \
            [packages.Package(None, DATA)]

        # Command to test
        self.cmd = osc_pkg.DeletePackage(self.app, None)

    @mock.patch('osc_lib.utils.get_item_properties')
    def test_package_delete(self, mock_util):
        arglist = ['fake1']
        verifylist = [('id', ['fake1'])]

        mock_util.return_value = ('1234', 'Core library',
                                  'io.murano', 'murano.io', '',
                                  'True', 'Library', '0.0.0'
                                  )

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        # Check that columns are correct
        self.assertEqual(COLUMNS, columns)

        # Check that data is correct
        expected_data = [('1234', 'Core library', 'io.murano',
                          'murano.io', '', 'True', 'Library', '0.0.0')]
        self.assertEqual(expected_data, data)


class TestPackageImport(TestPackage):
    def setUp(self):
        super(TestPackageImport, self).setUp()
        self.package_mock.filter.return_value = \
            [packages.Package(None, DATA)]

        # Command to test
        self.cmd = osc_pkg.ImportPackage(self.app, None)

    @mock.patch('muranoclient.common.utils.Package.from_file')
    def test_package_import(self, from_file):
        with tempfile.NamedTemporaryFile() as f:
            RESULT_PACKAGE = f.name
            categories = ['Cat1', 'Cat2 with space']

            pkg = make_pkg({'FullName': RESULT_PACKAGE})
            from_file.return_value = mc_utils.Package(mc_utils.File(pkg))

            arglist = [RESULT_PACKAGE, '--categories',
                       categories, '--is-public']
            parsed_args = self.check_parser(self.cmd, arglist, [])
            self.cmd.take_action(parsed_args)

            self.package_mock.create.assert_called_once_with({
                'categories': [categories],
                'is_public': True
            }, {RESULT_PACKAGE: mock.ANY},)

    def _test_conflict(self,
                       packages, from_file, raw_input_mock,
                       input_action, exists_action=''):
        packages.create = mock.MagicMock(
            side_effect=[common_exceptions.HTTPConflict("Conflict"), None])

        packages.filter.return_value = [mock.Mock(id='test_id')]

        raw_input_mock.return_value = input_action
        with tempfile.NamedTemporaryFile() as f:
            pkg = make_pkg({'FullName': f.name})
            from_file.return_value = mc_utils.Package(mc_utils.File(pkg))
            if exists_action:
                arglist = [f.name, '--exists-action', exists_action]
            else:
                arglist = [f.name]
            parsed_args = self.check_parser(self.cmd, arglist, [])
            self.cmd.take_action(parsed_args)

            return f.name

    @mock.patch('builtins.input')
    @mock.patch('muranoclient.common.utils.Package.from_file')
    def test_package_import_conflict_skip(self, from_file, raw_input_mock):

        name = self._test_conflict(
            self.package_mock,
            from_file,
            raw_input_mock,
            's',
        )

        self.package_mock.create.assert_called_once_with({
            'is_public': False,
        }, {name: mock.ANY},)

    @mock.patch('builtins.input')
    @mock.patch('muranoclient.common.utils.Package.from_file')
    def test_package_import_conflict_skip_ea(self, from_file, raw_input_mock):

        name = self._test_conflict(
            self.package_mock,
            from_file,
            raw_input_mock,
            '',
            exists_action='s',
        )

        self.package_mock.create.assert_called_once_with({
            'is_public': False,
        }, {name: mock.ANY},)
        self.assertFalse(raw_input_mock.called)

    @mock.patch('builtins.input')
    @mock.patch('muranoclient.common.utils.Package.from_file')
    def test_package_import_conflict_abort(self, from_file, raw_input_mock):

        self.assertRaises(SystemExit, self._test_conflict,
                          self.package_mock,
                          from_file,
                          raw_input_mock,
                          'a',
                          )

        self.package_mock.create.assert_called_once_with({
            'is_public': False,
        }, mock.ANY,)

    @mock.patch('builtins.input')
    @mock.patch('muranoclient.common.utils.Package.from_file')
    def test_package_import_conflict_abort_ea(self,
                                              from_file, raw_input_mock):

        self.assertRaises(SystemExit, self._test_conflict,
                          self.package_mock,
                          from_file,
                          raw_input_mock,
                          '',
                          exists_action='a',
                          )

        self.package_mock.create.assert_called_once_with({
            'is_public': False,
        }, mock.ANY,)
        self.assertFalse(raw_input_mock.called)

    @mock.patch('builtins.input')
    @mock.patch('muranoclient.common.utils.Package.from_file')
    def test_package_import_conflict_update(self, from_file, raw_input_mock):

        name = self._test_conflict(
            self.package_mock,
            from_file,
            raw_input_mock,
            'u',
        )

        self.assertEqual(2, self.package_mock.create.call_count)
        self.package_mock.delete.assert_called_once_with('test_id')

        self.package_mock.create.assert_has_calls(
            [
                mock.call({'is_public': False}, {name: mock.ANY},),
                mock.call({'is_public': False}, {name: mock.ANY},)
            ], any_order=True,
        )

    @mock.patch('builtins.input')
    @mock.patch('muranoclient.common.utils.Package.from_file')
    def test_package_import_conflict_update_ea(self,
                                               from_file, raw_input_mock):

        name = self._test_conflict(
            self.package_mock,
            from_file,
            raw_input_mock,
            '',
            exists_action='u',
        )

        self.assertEqual(2, self.package_mock.create.call_count)
        self.package_mock.delete.assert_called_once_with('test_id')

        self.package_mock.create.assert_has_calls(
            [
                mock.call({'is_public': False}, {name: mock.ANY},),
                mock.call({'is_public': False}, {name: mock.ANY},)
            ], any_order=True,
        )
        self.assertFalse(raw_input_mock.called)

    def _test_conflict_dep(self,
                           packages, from_file,
                           dep_exists_action=''):
        packages.create = mock.MagicMock(
            side_effect=[common_exceptions.HTTPConflict("Conflict"),
                         common_exceptions.HTTPConflict("Conflict"),
                         None])

        packages.filter.return_value = [mock.Mock(id='test_id')]

        pkg1 = make_pkg(
            {'FullName': 'first_app', 'Require': {'second_app': '1.0'}, })
        pkg2 = make_pkg({'FullName': 'second_app', })

        def side_effect(name):
            if 'first_app' in name:
                return mc_utils.Package(mc_utils.File(pkg1))
            if 'second_app' in name:
                return mc_utils.Package(mc_utils.File(pkg2))

        from_file.side_effect = side_effect

        arglist = ['first_app', '--exists-action', 's',
                   '--dep-exists-action', dep_exists_action]

        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)

    @mock.patch('muranoclient.common.utils.Package.from_file')
    def test_package_import_conflict_dep_skip_ea(self, from_file):
        self._test_conflict_dep(
            self.package_mock,
            from_file,
            dep_exists_action='s',
        )

        self.assertEqual(2, self.package_mock.create.call_count)
        self.package_mock.create.assert_has_calls(
            [
                mock.call({'is_public': False}, {'first_app': mock.ANY}),
                mock.call({'is_public': False}, {'second_app': mock.ANY}),
            ], any_order=True,
        )

    @mock.patch('muranoclient.common.utils.Package.from_file')
    def test_package_import_conflict_dep_abort_ea(self, from_file):
        self.assertRaises(SystemExit, self._test_conflict_dep,
                          self.package_mock,
                          from_file,
                          dep_exists_action='a',
                          )

        self.package_mock.create.assert_called_with({
            'is_public': False,
        }, {'second_app': mock.ANY},)

    @mock.patch('muranoclient.common.utils.Package.from_file')
    def test_package_import_conflict_dep_update_ea(self, from_file):
        self._test_conflict_dep(
            self.package_mock,
            from_file,
            dep_exists_action='u',
        )

        self.assertGreater(self.package_mock.create.call_count, 2)
        self.assertLess(self.package_mock.create.call_count, 5)

        self.assertTrue(self.package_mock.delete.called)

        self.package_mock.create.assert_has_calls(
            [
                mock.call({'is_public': False}, {'first_app': mock.ANY}),
                mock.call({'is_public': False}, {'second_app': mock.ANY}),
                mock.call({'is_public': False}, {'second_app': mock.ANY}),
            ], any_order=True,
        )

    @mock.patch('muranoclient.common.utils.Package.from_file')
    def test_package_import_no_categories(self, from_file):
        with tempfile.NamedTemporaryFile() as f:
            RESULT_PACKAGE = f.name
            pkg = make_pkg({'FullName': RESULT_PACKAGE})
            from_file.return_value = mc_utils.Package(mc_utils.File(pkg))

            arglist = [RESULT_PACKAGE]
            parsed_args = self.check_parser(self.cmd, arglist, [])
            self.cmd.take_action(parsed_args)

            self.package_mock.create.assert_called_once_with(
                {'is_public': False},
                {RESULT_PACKAGE: mock.ANY},
            )

    @requests_mock.mock()
    @mock.patch('muranoclient.common.utils.Package.from_file')
    def test_package_import_url(self, rm, from_file):
        filename = "http://127.0.0.1/test_package.zip"

        pkg = make_pkg({'FullName': 'test_package'})
        from_file.return_value = mc_utils.Package(mc_utils.File(pkg))

        rm.get(filename, body=make_pkg({'FullName': 'test_package'}))

        arglist = [filename]
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)

        self.package_mock.create.assert_called_once_with(
            {'is_public': False},
            {'test_package': mock.ANY},
        )

    @requests_mock.mock()
    @mock.patch('muranoclient.common.utils.Package.from_file')
    def test_package_import_by_name(self, rm, from_file):
        filename = "io.test.apps.test_application"
        murano_repo_url = "http://127.0.0.1"

        pkg = make_pkg({'FullName': filename})
        from_file.return_value = mc_utils.Package(mc_utils.File(pkg))

        rm.get(murano_repo_url + '/apps/' + filename + '.zip',
               body=make_pkg({'FullName': 'first_app'}))

        arglist = [filename, '--murano-repo-url', murano_repo_url]
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)

        self.assertTrue(self.package_mock.create.called)
        self.package_mock.create.assert_called_once_with(
            {'is_public': False},
            {filename: mock.ANY},
        )

    @requests_mock.mock()
    def test_package_import_multiple(self, rm):
        filename = ["io.test.apps.test_application",
                    "http://127.0.0.1/test_app2.zip", ]
        murano_repo_url = "http://127.0.0.1"

        rm.get(murano_repo_url + '/apps/' + filename[0] + '.zip',
               body=make_pkg({'FullName': 'first_app'}))

        rm.get(filename[1],
               body=make_pkg({'FullName': 'second_app'}))

        arglist = [filename[0], filename[1],
                   '--murano-repo-url', murano_repo_url]
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)

        self.assertEqual(2, self.package_mock.create.call_count)

        self.package_mock.create.assert_has_calls(
            [
                mock.call({'is_public': False}, {'first_app': mock.ANY}),
                mock.call({'is_public': False}, {'second_app': mock.ANY}),
            ], any_order=True,
        )


class TestBundleImport(TestPackage):
    def setUp(self):
        super(TestBundleImport, self).setUp()

        # Command to test
        self.cmd = osc_pkg.ImportBundle(self.app, None)

    @requests_mock.mock()
    def test_import_bundle_by_name(self, m):
        """Asserts bundle import calls packages create once for each pkg."""
        pkg1 = make_pkg({'FullName': 'first_app'})
        pkg2 = make_pkg({'FullName': 'second_app'})

        murano_repo_url = "http://127.0.0.1"

        m.get(murano_repo_url + '/apps/first_app.zip', body=pkg1)
        m.get(murano_repo_url + '/apps/second_app.1.0.zip',
              body=pkg2)
        s = io.StringIO()
        bundle_contents = {'Packages': [
            {'Name': 'first_app'},
            {'Name': 'second_app', 'Version': '1.0'}
        ]}
        json.dump(bundle_contents, s)
        s = io.BytesIO(s.getvalue().encode('ascii'))

        m.get(murano_repo_url + '/bundles/test_bundle.bundle',
              body=s)

        arglist = ["test_bundle", '--murano-repo-url', murano_repo_url]
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)

        self.package_mock.create.assert_has_calls(
            [
                mock.call({'is_public': False}, {'first_app': mock.ANY}),
                mock.call({'is_public': False}, {'second_app': mock.ANY}),
            ], any_order=True,
        )

    @requests_mock.mock()
    def test_import_bundle_wrong_url(self, m):
        url = 'http://127.0.0.2/test_bundle.bundle'
        m.get(url, status_code=404)

        arglist = [url]
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)

        self.assertFalse(self.package_mock.packages.create.called)

    @requests_mock.mock()
    def test_import_bundle_no_bundle(self, m):
        url = 'http://127.0.0.1/bundles/test_bundle.bundle'
        m.get(url, status_code=404)

        arglist = ["test_bundle"]
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)

        self.assertFalse(self.package_mock.packages.create.called)

    @requests_mock.mock()
    def test_import_bundle_by_url(self, m):
        """Asserts bundle import calls packages create once for each pkg."""
        pkg1 = make_pkg({'FullName': 'first_app'})
        pkg2 = make_pkg({'FullName': 'second_app'})

        murano_repo_url = 'http://127.0.0.1'
        m.get(murano_repo_url + '/apps/first_app.zip', body=pkg1)
        m.get(murano_repo_url + '/apps/second_app.1.0.zip',
              body=pkg2)
        s = io.StringIO()
        bundle_contents = {'Packages': [
            {'Name': 'first_app'},
            {'Name': 'second_app', 'Version': '1.0'}
        ]}
        json.dump(bundle_contents, s)
        s = io.BytesIO(s.getvalue().encode('ascii'))

        url = 'http://127.0.0.2/test_bundle.bundle'
        m.get(url, body=s)

        arglist = [url, '--murano-repo-url', murano_repo_url]
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)

        self.package_mock.create.assert_has_calls(
            [
                mock.call({'is_public': False}, {'first_app': mock.ANY}),
                mock.call({'is_public': False}, {'second_app': mock.ANY}),
            ], any_order=True,
        )

    @requests_mock.mock()
    def test_import_local_bundle(self, m):
        """Asserts local bundles are first searched locally."""
        tmp_dir = tempfile.mkdtemp()
        bundle_file = os.path.join(tmp_dir, 'bundle.bundle')
        with open(os.path.join(tmp_dir, 'bundle.bundle'), 'w') as f:

            bundle_contents = {'Packages': [
                {'Name': 'first_app'},
                {'Name': 'second_app', 'Version': '1.0'}
            ]}
            json.dump(bundle_contents, f)

        pkg1 = make_pkg({'FullName': 'first_app',
                         'Require': {'third_app': None}})
        pkg2 = make_pkg({'FullName': 'second_app'})
        pkg3 = make_pkg({'FullName': 'third_app'})
        with open(os.path.join(tmp_dir, 'first_app'), 'wb') as f:
            f.write(pkg1.read())
        with open(os.path.join(tmp_dir, 'third_app'), 'wb') as f:
            f.write(pkg3.read())

        murano_repo_url = "http://127.0.0.1"
        m.get(murano_repo_url + '/apps/first_app.zip',
              status_code=404)
        m.get(murano_repo_url + '/apps/second_app.1.0.zip',
              body=pkg2)
        m.get(murano_repo_url + '/apps/third_app.zip',
              status_code=404)

        arglist = [bundle_file, '--murano-repo-url', murano_repo_url]
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)

        self.package_mock.create.assert_has_calls(
            [
                mock.call({'is_public': False}, {'first_app': mock.ANY}),
                mock.call({'is_public': False}, {'second_app': mock.ANY}),
                mock.call({'is_public': False}, {'third_app': mock.ANY}),
            ], any_order=True,
        )
        shutil.rmtree(tmp_dir)


class TestShowPackage(TestPackage):
    def setUp(self):
        super(TestShowPackage, self).setUp()

        # Command to test
        self.cmd = osc_pkg.ShowPackage(self.app, None)

    def test_package_show(self):
        arglist = ['fake']
        verifylist = [('id', 'fake')]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        # Check that columns are correct
        expected_columns = ('categories', 'class_definitions', 'description',
                            'enabled', 'fully_qualified_name', 'id',
                            'is_public', 'name', 'owner_id', 'tags', 'type')
        self.assertEqual(expected_columns, columns)

        self.package_mock.get.assert_called_with('fake')


class TestUpdatePackage(TestPackage):
    def setUp(self):
        super(TestUpdatePackage, self).setUp()

        self.package_mock.update.return_value = \
            (mock.MagicMock(), mock.MagicMock())

        # Command to test
        self.cmd = osc_pkg.UpdatePackage(self.app, None)

    def test_package_update(self):
        arglist = ['123', '--is-public', 'true']
        verifylist = [('id', '123'), ('is_public', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        self.package_mock.update.assert_called_with('123', {'is_public': True})

        arglist = ['123', '--enabled', 'true']
        verifylist = [('id', '123'), ('enabled', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        self.package_mock.update.assert_called_with('123', {'enabled': True})

        arglist = ['123', '--name', 'foo', '--description', 'bar']
        verifylist = [('id', '123'), ('name', 'foo'), ('description', 'bar')]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        self.package_mock.update.assert_called_with(
            '123', {'name': 'foo', 'description': 'bar'})

        arglist = ['123', '--tags', 'foo']
        verifylist = [('id', '123'), ('tags', ['foo'])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        self.package_mock.update.assert_called_with(
            '123', {'tags': ['foo']})


class TestDownloadPackage(TestPackage):
    def setUp(self):
        super(TestDownloadPackage, self).setUp()

        self.package_mock.download.return_value = \
            b'This is a fake package buffer'

        # Command to test
        self.cmd = osc_pkg.DownloadPackage(self.app, None)

    def test_package_download(self):
        arglist = ['1234', '/tmp/foo.zip']
        verifylist = [('id', '1234'), ('filename', '/tmp/foo.zip')]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        self.package_mock.download.assert_called_with('1234')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Application-catalog v1 package action implementation"""

import collections
import functools
import itertools
import os
import shutil
import sys
import tempfile
import zipfile

from osc_lib.command import command
from osc_lib import exceptions as exc
from osc_lib import utils
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import strutils

from muranoclient.apiclient import exceptions
from muranoclient.common import exceptions as common_exceptions
from muranoclient.common import utils as murano_utils
from muranoclient.v1.package_creator import hot_package
from muranoclient.v1.package_creator import mpl_package


LOG = logging.getLogger(__name__)

DEFAULT_REPO_URL = "http://apps.openstack.org/api/v1/murano_repo/liberty/"

_bool_from_str_strict = functools.partial(
    strutils.bool_from_string, strict=True)


class CreatePackage(command.Command):
    """Create an application package."""

    def get_parser(self, prog_name):
        parser = super(CreatePackage, self).get_parser(prog_name)
        parser.add_argument(
            '-t', '--template',
            metavar='<HEAT_TEMPLATE>',
            help=("Path to the Heat template to import as "
                  "an Application Definition."),
        )
        parser.add_argument(
            '-c', '--classes-dir',
            metavar='<CLASSES_DIRECTORY>',
            help=("Path to the directory containing application classes."),
        )
        parser.add_argument(
            '-r', '--resources-dir',
            metavar='<RESOURCES_DIRECTORY>',
            help=("Path to the directory containing application resources."),
        )
        parser.add_argument(
            '-n', '--name',
            metavar='<DISPLAY_NAME>',
            help=("Display name of the Application in Catalog."),
        )
        parser.add_argument(
            '-f', '--full-name',
            metavar='<full-name>',
            help=("Fully-qualified name of the Application in Catalog."),
        )
        parser.add_argument(
            '-a', '--author',
            metavar='<AUTHOR>',
            help=("Name of the publisher."),
        )
        parser.add_argument(
            '--tags',
            metavar='<TAG1 TAG2>',
            nargs='*',
            help=("A list of keywords connected to the application."),
        )
        parser.add_argument(
            '-d', '--description',
            metavar='<DESCRIPTION>',
            help=("Detailed description for the Application in Catalog."),
        )
        parser.add_argument(
            '-o', '--output',
            metavar='<PACKAGE_NAME>',
            help=("The name of the output file archive to save locally."),
        )
        parser.add_argument(
            '-u', '--ui',
            metavar='<UI_DEFINITION>',
            help=("Dynamic UI form definition."),
        )
        parser.add_argument(
            '--type',
            metavar='<TYPE>',
            help=("Package type. Possible values: Application or Library."),
        )
        parser.add_argument(
            '-l', '--logo',
            metavar='<LOGO>',
            help=("Path to the package logo."),
        )

        return parser

    def take_action(self, parsed_args):
        LOG.debug("take_action({0})".format(parsed_args))
        parsed_args.os_username = os.getenv('OS_USERNAME')

        def _make_archive(archive_name, path):
            zip_file = zipfile.ZipFile(archive_name, 'w')
            for root, dirs, files in os.walk(path):
                for f in files:
                    zip_file.write(os.path.join(root, f),
                                   arcname=os.path.join(
                                   os.path.relpath(root, path), f))

        if parsed_args.template and parsed_args.classes_dir:
            raise exc.CommandError(
                "Provide --template for a HOT-based package, OR"
                " --classes-dir for a MuranoPL-based package")
        if not parsed_args.template and not parsed_args.classes_dir:
            raise exc.CommandError(
                "Provide --template for a HOT-based package, OR at least"
                " --classes-dir for a MuranoPL-based package")
        directory_path = None
        try:
            archive_name = parsed_args.output if parsed_args.output else None
            if parsed_args.template:
                directory_path = hot_package.prepare_package(parsed_args)
                if not archive_name:
                    archive_name = os.path.basename(parsed_args.template)
                    archive_name = os.path.splitext(archive_name)[0] + ".zip"
            else:
                directory_path = mpl_package.prepare_package(parsed_args)
                if not archive_name:
                    archive_name = tempfile.mkstemp(
                        prefix="murano_", dir=os.getcwd())[1] + ".zip"

            _make_archive(archive_name, directory_path)
            print("Application package is available at " +
                  os.path.abspath(archive_name))
        finally:
            if directory_path:
                shutil.rmtree(directory_path)


class ListPackages(command.Lister):
    """List available packages."""

    def get_parser(self, prog_name):
        parser = super(ListPackages, self).get_parser(prog_name)
        parser.add_argument(
            "--limit",
            type=int,
            default=0,
            help='Show limited number of packages'
        )
        parser.add_argument(
            "--marker",
            default='',
            help='Show packages starting from package with id excluding it'
        )
        parser.add_argument(
            "--include-disabled",
            default=False,
            action="store_true"
        )
        parser.add_argument(
            "--owned",
            default=False,
            action="store_true"
        )
        parser.add_argument(
            '--search',
            metavar='<SEARCH_KEYS>',
            dest='search',
            required=False,
            help='Show packages, that match search keys fuzzily'
        )
        parser.add_argument(
            '--name',
            metavar='<PACKAGE_NAME>',
            dest='name',
            required=False,
            help='Show packages, whose name match parameter exactly'
        )
        parser.add_argument(
            '--fqn',
            metavar="<PACKAGE_FULLY_QUALIFIED_NAME>",
            dest='fqn',
            required=False,
            help='Show packages, '
                 'whose fully qualified name match parameter exactly'
        )
        parser.add_argument(
            '--type',
            metavar='<PACKAGE_TYPE>',
            dest='type',
            required=False,
            help='Show packages, whose type match parameter exactly'
        )
        parser.add_argument(
            '--category',
            metavar='<PACKAGE_CATEGORY>',
            dest='category',
            required=False,
            help='Show packages, whose categories include parameter'
        )
        parser.add_argument(
            '--class_name',
            metavar='<PACKAGE_CLASS_NAME>',
            dest='class_name',
            required=False,
            help='Show packages, whose class name match parameter exactly'
        )
        parser.add_argument(
            '--tag',
            metavar='<PACKAGE_TAG>',
            dest='tag',
            required=False,
            help='Show packages, whose tags include parameter'
        )

        return parser

    def take_action(self, parsed_args):
        LOG.debug("take_action({0})".format(parsed_args))
        client = self.app.client_manager.application_catalog
        filter_args = {
            "include_disabled": getattr(parsed_args,
                                        'include_disabled', False),
            "owned": getattr(parsed_args, 'owned', False),
        }
        if parsed_args:
            if parsed_args.limit < 0:
                raise exceptions.CommandError(
                    '--limit parameter must be non-negative')
            if parsed_args.limit != 0:
                filter_args['limit'] = parsed_args.limit
            if parsed_args.marker:
                filter_args['marker'] = parsed_args.marker
            if parsed_args.search:
                filter_args['search'] = parsed_args.search
            if parsed_args.name:
                filter_args['name'] = parsed_args.name
            if parsed_args.fqn:
                filter_args['fqn'] = parsed_args.fqn
            if parsed_args.type:
                filter_args['type'] = parsed_args.type
            if parsed_args.category:
                filter_args['category'] = parsed_args.category
            if parsed_args.class_name:
                filter_args['class_name'] = parsed_args.class_name
            if parsed_args.tag:
                filter_args['tag'] = parsed_args.tag

        data = client.packages.filter(**filter_args)

        columns = ('id', 'name', 'fully_qualified_name', 'author', 'active',
                   'is public', 'type', 'version')
        column_headers = [c.capitalize() for c in columns]
        if not parsed_args or parsed_args.limit == 0:
            return (
                column_headers,
                list(utils.get_item_properties(
                    s,
                    columns,
                ) for s in data)
            )
        else:
            return (
                column_headers,
                list(utils.get_item_properties(
                    s,
                    columns,
                ) for s in itertools.islice(data, parsed_args.limit))
            )


class DeletePackage(command.Lister):
    """Delete a package."""

    def get_parser(self, prog_name):
        parser = super(DeletePackage, self).get_parser(prog_name)
        parser.add_argument(
            'id',
            metavar="<ID>",
            nargs="+",
            help="Package ID to delete.",
        )

        return parser

    def take_action(self, parsed_args):
        LOG.debug("take_action({0})".format(parsed_args))

        client = self.app.client_manager.application_catalog

        failure_count = 0
        for package_id in parsed_args.id:
            try:
                client.packages.delete(package_id)
            except exceptions.NotFound:
                failure_count += 1
                print("Failed to delete '{0}'; package not found".
                      format(package_id))

        if failure_count == len(parsed_args.id):
            raise exceptions.CommandError("Unable to find and delete any of "
                                          "the specified packages.")
        data = client.packages.filter()

        columns = ('id', 'name', 'fully_qualified_name', 'author', 'active',
                   'is public', 'type', 'version')
        column_headers = [c.capitalize() for c in columns]

        return (
            column_headers,
            list(utils.get_item_properties(
                s,
                columns,
            ) for s in data)
        )


def _handle_package_exists(mc, data, package, exists_action):
    name = package.manifest['FullName']
    version = package.manifest.get('Version', '0')
    while True:
        print("Importing package {0}".format(name))
        try:
            return mc.packages.create(data, {name: package.file()})
        except common_exceptions.HTTPConflict:
            print("Importing package {0} failed. Package with the same"
                  " name/classes is already registered.".format(name))
            allowed_results = ['s', 'u', 'a']
            res = exists_action
            if not res:
                while True:
                    print("What do you want to do? (s)kip, (u)pdate, (a)bort")
                    res = input()
                    if res in allowed_results:
                        break
            if res == 's':
                print("Skipping.")
                return None
            elif res == 'a':
                print("Exiting.")
                sys.exit()
            elif res == 'u':
                pkgs = list(mc.packages.filter(fqn=name, version=version,
                                               owned=True))
                if not pkgs:
                    msg = (
                        "Got a conflict response, but could not find the "
                        "package '{0}' in the current tenant.\nThis probably "
                        "means the conflicting package is in another tenant.\n"
                        "Please delete it manually."
                    ).format(name)
                    raise exceptions.CommandError(msg)
                elif len(pkgs) > 1:
                    msg = (
                        "Got {0} packages with name '{1}'.\nI do not trust "
                        "myself, please delete the package manually."
                    ).format(len(pkgs), name)
                    raise exceptions.CommandError(msg)
                print("Deleting package {0}({1})".format(name, pkgs[0].id))
                mc.packages.delete(pkgs[0].id)
                continue


class ImportPackage(command.Lister):
    """Import a package."""

    def get_parser(self, prog_name):
        parser = super(ImportPackage, self).get_parser(prog_name)
        parser.add_argument(
            'filename',
            metavar='<FILE>',
            nargs='+',
            help='URL of the murano zip package, FQPN, path to zip package'
                 ' or path to directory with package.'
        )
        parser.add_argument(
            '--categories',
            metavar='<CATEGORY>',
            nargs='*',
            help='Category list to attach.',
        )
        parser.add_argument(
            '--is-public',
            action='store_true',
            default=False,
            help="Make the package available for users from other tenants.",
        )
        parser.add_argument(
            '--package-version',
            default='',
            help='Version of the package to use from repository '
                 '(ignored when importing with multiple packages).'
        )
        parser.add_argument(
            '--exists-action',
            default='',
            choices=['a', 's', 'u'],
            help='Default action when a package already exists: '
                 '(s)kip, (u)pdate, (a)bort.'
        )
        parser.add_argument(
            '--dep-exists-action',
            default='',
            choices=['a', 's', 'u'],
            help='Default action when a dependency package already exists: '
                 '(s)kip, (u)pdate, (a)bort.'
        )
        parser.add_argument('--murano-repo-url',
                            default=murano_utils.env(
                                'MURANO_REPO_URL',
                                default=DEFAULT_REPO_URL),
                            help=('Defaults to env[MURANO_REPO_URL] '
                                  'or {0}'.format(DEFAULT_REPO_URL)))

        return parser

    def take_action(self, parsed_args):
        LOG.debug("take_action({0})".format(parsed_args))

        client = self.app.client_manager.application_catalog

        data = {"is_public": parsed_args.is_public}
        version = parsed_args.package_version
        if version and len(parsed_args.filename) >= 2:
            print("Requested to import more than one package, "
                  "ignoring version.")
            version = ''

        if parsed_args.categories:
            data["categories"] = parsed_args.categories

        total_reqs = collections.OrderedDict()
        main_packages_names = []
        for filename in parsed_args.filename:
            if os.path.isfile(filename) or os.path.isdir(filename):
                _file = filename
            else:
                print("Package file '{0}' does not exist, attempting to "
                      "download".format(filename))
                _file = murano_utils.to_url(
                    filename,
                    version=version,
                    base_url=parsed_args.murano_repo_url,
                    extension='.zip',
                    path='apps/',
                )
            try:
                package = murano_utils.Package.from_file(_file)
            except Exception as e:
                print("Failed to create package for '{0}', reason: {1}".format(
                    filename, e))
                continue
            total_reqs.update(
                package.requirements(base_url=parsed_args.murano_repo_url))
            main_packages_names.append(package.manifest['FullName'])

        imported_list = []

        dep_exists_action = parsed_args.dep_exists_action
        if dep_exists_action == '':
            dep_exists_action = parsed_args.exists_action

        for name, package in iter(total_reqs.items()):
            image_specs = package.images()
            if image_specs:
                print("Inspecting required images")
                try:
                    imgs = murano_utils.ensure_images(
                        glance_client=client.glance_client,
                        image_specs=image_specs,
                        base_url=parsed_args.murano_repo_url,
                        is_package_public=parsed_args.is_public)
                    for img in imgs:
                        print("Added {0}, {1} image".format(
                            img['name'], img['id']))
                except Exception as e:
                    print("Error {0} occurred while installing "
                          "images for {1}".format(e, name))

            if name in main_packages_names:
                exists_action = parsed_args.exists_action
            else:
                exists_action = dep_exists_action
            try:
                imported_package = _handle_package_exists(
                    client, data, package, exists_action)
                if imported_package:
                    imported_list.append(imported_package)
            except Exception as e:
                print("Error {0} occurred while installing package {1}".format(
                    e, name))

        columns = ('id', 'name', 'fully_qualified_name', 'author', 'active',
                   'is public', 'type', 'version')
        column_headers = [c.capitalize() for c in columns]

        return (
            column_headers,
            list(utils.get_item_properties(
                s,
                columns,
            ) for s in imported_list)
        )


class ImportBundle(command.Lister):
    """Import a bundle."""

    def get_parser(self, prog_name):
        parser = super(ImportBundle, self).get_parser(prog_name)
        parser.add_argument(
            'filename',
            metavar='<FILE>',
            nargs='+',
            help='Bundle URL, bundle name, or path to the bundle file.'
        )
        parser.add_argument(
            '--is-public',
            action='store_true',
            default=False,
            help="Make the package available for users from other tenants.",
        )
        parser.add_argument(
            '--exists-action',
            default='',
            choices=['a', 's', 'u'],
            help='Default action when a package already exists: '
                 '(s)kip, (u)pdate, (a)bort.'
        )
        parser.add_argument('--murano-repo-url',
                            default=murano_utils.env(
                                'MURANO_REPO_URL',
                                default=DEFAULT_REPO_URL),
                            help=('Defaults to env[MURANO_REPO_URL] '
                                  'or {0}'.format(DEFAULT_REPO_URL)))

        return parser

    def take_action(self, parsed_args):

        LOG.debug("take_action({0})".format(parsed_args))

        client = self.app.client_manager.application_catalog

        total_reqs = collections.OrderedDict()
        for filename in parsed_args.filename:
            local_path = None
            if os.path.isfile(filename):
                _file = filename
                local_path = os.path.dirname(os.path.abspath(filename))
            else:
                print("Bundle file '{0}' does not exist, attempting "
                      "to download".format(filename))
                _file = murano_utils.to_url(
                    filename,
                    base_url=parsed_args.murano_repo_url,
                    path='bundles/',
                    extension='.bundle',
                )

            try:
                bundle_file = murano_utils.Bundle.from_file(_file)
            except Exception as e:
                print("Failed to create bundle for '{0}', reason: {1}".format(
                    filename, e))
                continue

            data = {"is_public": parsed_args.is_public}

            for package in bundle_file.packages(
                    base_url=parsed_args.murano_repo_url, path=local_path):

                requirements = package.requirements(
                    base_url=parsed_args.murano_repo_url,
                    path=local_path,
                )
                total_reqs.update(requirements)

        imported_list = []

        for name, dep_package in total_reqs.items():
            image_specs = dep_package.images()
            if image_specs:
                print("Inspecting required images")
                try:
                    imgs = parsed_args.ensure_images(
                        glance_client=client.glance_client,
                        image_specs=image_specs,
                        base_url=parsed_args.murano_repo_url,
                        local_path=local_path,
                        is_package_public=parsed_args.is_public)
                    for img in imgs:
                        print("Added {0}, {1} image".format(
                            img['name'], img['id']))
                except Exception as e:
                    print("Error {0} occurred while installing "
                          "images for {1}".format(e, name))
            try:
                imported_package = _handle_package_exists(
                    client, data, dep_package, parsed_args.exists_action)
                if imported_package:
                    imported_list.append(imported_package)
            except exceptions.CommandError:
                raise
            except Exception as e:
                print("Error {0} occurred while "
                      "installing package {1}".format(e, name))

        columns = ('id', 'name', 'fully_qualified_name', 'author', 'active',
                   'is public', 'type', 'version')
        column_headers = [c.capitalize() for c in columns]

        return (
            column_headers,
            list(utils.get_item_properties(
                s,
                columns,
            ) for s in imported_list)
        )


class ShowPackage(command.ShowOne):
    """Display details for a package."""

    def get_parser(self, prog_name):
        parser = super(ShowPackage, self).get_parser(prog_name)
        parser.add_argument(
            "id",
            metavar="<ID>",
            help=("Package ID to show."),
        )

        return parser

    def take_action(self, parsed_args):
        LOG.debug("take_action({0})".format(parsed_args))
        client = self.app.client_manager.application_catalog

        try:
            package = client.packages.get(parsed_args.id)
        except common_exceptions.HTTPNotFound:
            raise exceptions.CommandError("Package with id %s not "
                                          "found" % parsed_args.id)
        else:
            to_display = dict(
                id=package.id,
                type=package.type,
                owner_id=package.owner_id,
                name=package.name,
                fully_qualified_name=package.fully_qualified_name,
                is_public=package.is_public,
                enabled=package.enabled,
                class_definitions=jsonutils.dumps(package.class_definitions,
                                                  indent=2),
                categories=jsonutils.dumps(package.categories, indent=2),
                tags=jsonutils.dumps(package.tags, indent=2),
                description=package.description
            )

        return self.dict2columns(to_display)


class UpdatePackage(command.ShowOne):
    """Update an existing package."""

    def get_parser(self, prog_name):
        parser = super(UpdatePackage, self).get_parser(prog_name)
        parser.add_argument(
            'id',
            metavar="<ID>",
            help="Package ID to update.",
        )
        parser.add_argument(
            '--is-public',
            type=_bool_from_str_strict,
            metavar="{true|false}",
            help="Make package available to users from other tenants.",
        )
        parser.add_argument(
            '--enabled',
            type=_bool_from_str_strict,
            metavar="{true|false}",
            help="Make package active and available for deployments.",
        )
        parser.add_argument(
            '--name',
            default=None,
            help="New name for the package.",
        )
        parser.add_argument(
            '--description',
            default=None,
            help="New package description.",
        )
        parser.add_argument(
            '--tags',
            metavar='<TAG>', nargs='*',
            default=None,
            help="A list of keywords connected to the application.",
        )

        return parser

    def take_action(self, parsed_args):
        LOG.debug("take_action({0})".format(parsed_args))
        client = self.app.client_manager.application_catalog
        data = {}
        parameters = ('is_public', 'enabled',
                      'name', 'description',
                      'tags')
        for parameter in parameters:
            param_value = getattr(parsed_args, parameter, None)
            if param_value is not None:
                data[parameter] = param_value

        _, package = client.packages.update(parsed_args.id, data)

        to_display = dict(
            id=package["id"],
            type=package["type"],
            owner_id=package["owner_id"],
            name=package["name"],
            fully_qualified_name=package["fully_qualified_name"],
            is_public=package["is_public"],
            enabled=package["enabled"],
            class_definitions=jsonutils.dumps(package["class_definitions"],
                                              indent=2),
            categories=jsonutils.dumps(package["categories"], indent=2),
            tags=jsonutils.dumps(package["tags"], indent=2),
            description=package["description"]
        )

        return self.dict2columns(to_display)


class DownloadPackage(command.Command):
    """Download a package to a filename or stdout."""

    def get_parser(self, prog_name):
        parser = super(DownloadPackage, self).get_parser(prog_name)
        parser.add_argument(
            "id",
            metavar="<ID>",
            help=("Package ID to download."),
        )
        parser.add_argument(
            "filename",
            metavar="file", nargs="?",
            help=("Filename to save package to. If it is not "
                  "specified and there is no stdout redirection "
                  "the package won't be saved."),
        )

        return parser

    def take_action(self, parsed_args):
        LOG.debug("take_action({0})".format(parsed_args))
        client = self.app.client_manager.application_catalog

        def download_to_fh(package_id, fh):
            try:
                fh.write(client.packages.download(package_id))
            except common_exceptions.HTTPNotFound:
                raise exceptions.CommandError("Package with id %s not "
                                              "found" % parsed_args.id)

        if parsed_args.filename:
            with open(parsed_args.filename, 'wb') as fh:
                download_to_fh(parsed_args.id, fh)
                print("Package downloaded to %s" % parsed_args.filename)
        elif not sys.stdout.isatty():
            download_to_fh(parsed_args.id, sys.stdout)
        else:
            msg = ("No stdout redirection or local file specified for "
                   "downloaded package. Please specify a local file to "
                   "save downloaded package or redirect output to "
                   "another source.")
            raise exceptions.CommandError(msg)



# Copyright 2012 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re
import sys


# TODO(sjmc7): This module is likely redundant because it's replaced
# by openstack.common.apiclient; should be removed
class BaseException(Exception):
    """An error occurred."""
    def __init__(self, message=None):
        self.message = message

    def __str__(self):
        return self.message or self.__class__.__doc__


class InvalidEndpoint(BaseException):
    """The provided endpoint is invalid."""


class CommunicationError(BaseException):
    """Unable to communicate with server."""


class ClientException(Exception):
    """DEPRECATED!"""


class HTTPException(ClientException):
    """Base exception for all HTTP-derived exceptions."""
    code = 'N/A'

    def __init__(self, details=None):
        self.details = details or self.__class__.__name__

    def __str__(self):
        return "%s (HTTP %s)" % (self.details, self.code)


class HTTPMultipleChoices(HTTPException):
    code = 300

    def __str__(self):
        self.details = ("Requested version of Application Catalog API is not "
                        "available.")
        return "%s (HTTP %s) %s" % (self.__class__.__name__, self.code,
                                    self.details)


class BadRequest(HTTPException):
    """DEPRECATED!"""
    code = 400


class HTTPBadRequest(BadRequest):
    pass


class Unauthorized(HTTPException):
    """DEPRECATED!"""
    code = 401


class HTTPUnauthorized(Unauthorized):
    pass


class Forbidden(HTTPException):
    """DEPRECATED!"""
    code = 403


class HTTPForbidden(Forbidden):
    pass


class NotFound(HTTPException):
    """DEPRECATED!"""
    code = 404


class HTTPNotFound(NotFound):
    pass


class HTTPMethodNotAllowed(HTTPException):
    code = 405


class Conflict(HTTPException):
    """DEPRECATED!"""
    code = 409


class HTTPConflict(Conflict):
    pass


class OverLimit(HTTPException):
    """DEPRECATED!"""
    code = 413


class HTTPOverLimit(OverLimit):
    pass


class HTTPInternalServerError(HTTPException):
    code = 500


class HTTPNotImplemented(HTTPException):
    code = 501


class HTTPBadGateway(HTTPException):
    code = 502


class ServiceUnavailable(HTTPException):
    """DEPRECATED!"""
    code = 503


class HTTPServiceUnavailable(ServiceUnavailable):
    pass


# NOTE(bcwaldon): Build a mapping of HTTP codes to corresponding exception
# classes
_code_map = {}
for obj_name in dir(sys.modules[__name__]):
    if obj_name.startswith('HTTP'):
        obj = getattr(sys.modules[__name__], obj_name)
        _code_map[obj.code] = obj


def from_response(response):
    """Return an instance of an HTTPException based on httplib response."""
    cls = _code_map.get(response.status_code, HTTPException)
    body = response.content
    if body and response.headers['content-type'].\
       lower().startswith("application/json"):
        # Iterate over the nested objects and retrieve the "message" attribute.
        messages = [obj.get('message') for obj in response.json().values()]
        # Join all of the messages together nicely and filter out any objects
        # that don't have a "message" attr.
        details = '\n'.join(i for i in messages if i is not None)
        return cls(details=details)
    elif body and \
            response.headers['content-type'].lower().startswith("text/html"):
        # Split the lines, strip whitespace and inline HTML from the response.
        details = [re.sub(r'<.+?>', '', i.strip())
                   for i in response.text.splitlines()]
        details = [i for i in details if i]
        # Remove duplicates from the list.
        details_seen = set()
        details_temp = []
        for i in details:
            if i not in details_seen:
                details_temp.append(i)
                details_seen.add(i)
        # Return joined string separated by colons.
        details = ': '.join(details_temp)
        return cls(details=details)
    elif body:
        details = body.replace('\n\n', '\n')
        return cls(details=details)

    return cls()


def from_code(code):
    cls = _code_map.get(code, HTTPException)
    return cls()


class NoTokenLookupException(Exception):
    """DEPRECATED!"""
    pass


class EndpointNotFound(Exception):
    """DEPRECATED!"""
    pass


class SSLConfigurationError(BaseException):
    pass


class SSLCertificateError(BaseException):
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from muranoclient.osc.v1 import category as osc_category
from muranoclient.tests.unit.osc.v1 import fakes
from muranoclient.v1 import categories as api_category
from muranoclient.v1 import packages as api_packages

CATEGORY_INFO = {'id': 'xyz123',
                 'name': 'fake1',
                 'packages': [{'name': 'package1'}, {'name': 'package2'}]}


class TestCategory(fakes.TestApplicationCatalog):
    def setUp(self):
        super(TestCategory, self).setUp()
        self.category_mock = self.app.client_manager.application_catalog.\
            categories
        self.category_mock.reset_mock()
        self.packages_mock = \
            self.app.client_manager.application_catalog.packages


class TestListCategories(TestCategory):
    def setUp(self):
        super(TestListCategories, self).setUp()
        self.category_mock.list.return_value = [api_category.Category(None,
                                                CATEGORY_INFO)]

        # Command to test
        self.cmd = osc_category.ListCategories(self.app, None)

    @mock.patch('osc_lib.utils.get_item_properties')
    def test_category_list(self, mock_util):
        mock_util.return_value = ('xyz123', 'fake1')

        columns, data = self.cmd.take_action(parsed_args=None)

        # Check that columns are correct
        expected_columns = ['ID', 'Name']
        self.assertEqual(expected_columns, columns)

        # Check that data is correct
        expected_data = [('xyz123', 'fake1')]
        self.assertEqual(expected_data, data)


class TestShowCategory(TestCategory):
    def setUp(self):
        super(TestShowCategory, self).setUp()
        self.category_mock.get.return_value = api_category.\
            Category(None, CATEGORY_INFO)

        self.packages_mock.filter.return_value = [
            api_packages.Package(None, pkg_info) for pkg_info in CATEGORY_INFO[
                'packages']
        ]

        # Command to test
        self.cmd = osc_category.ShowCategory(self.app, None)

    @mock.patch('textwrap.wrap')
    def test_category_show(self, mock_wrap):
        arglist = ['xyz123']
        verifylist = [('id', 'xyz123')]

        mock_wrap.return_value = ['package1, package2']

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        # Check that columns are correct
        expected_columns = ('id', 'name', 'packages')
        self.assertEqual(expected_columns, columns)

        # Check that data is correct
        expected_data = ('xyz123', 'fake1', 'package1, package2')
        self.assertEqual(expected_data, data)


class TestCreateCategory(TestCategory):
    def setUp(self):
        super(TestCreateCategory, self).setUp()
        self.category_mock.add.return_value = [api_category.Category(None,
                                               CATEGORY_INFO)]

        # Command to test
        self.cmd = osc_category.CreateCategory(self.app, None)

    @mock.patch('osc_lib.utils.get_item_properties')
    def test_category_list(self, mock_util):
        arglist = ['fake1']
        verifylist = [('name', 'fake1')]

        mock_util.return_value = ('xyz123', 'fake1')

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        # Check that columns are correct
        expected_columns = ['ID', 'Name']
        self.assertEqual(expected_columns, columns)

        # Check that data is correct
        expected_data = [('xyz123', 'fake1')]
        self.assertEqual(expected_data, data)


class TestDeleteCategory(TestCategory):
    def setUp(self):
        super(TestDeleteCategory, self).setUp()
        self.category_mock.delete.return_value = None
        self.category_mock.list.return_value = [api_category.Category(None,
                                                CATEGORY_INFO)]

        # Command to test
        self.cmd = osc_category.DeleteCategory(self.app, None)

    @mock.patch('osc_lib.utils.get_item_properties')
    def test_category_list(self, mock_util):
        arglist = ['abc123', '123abc']
        verifylist = [('id', ['abc123', '123abc'])]

        mock_util.return_value = ('xyz123', 'fake1')

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        # Check that columns are correct
        expected_columns = ['ID', 'Name']
        self.assertEqual(expected_columns, columns)

        # Check that data is correct
        expected_data = [('xyz123', 'fake1')]
        self.assertEqual(expected_data, data)



# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from glanceclient import exc


class ArtifactType(object):
    generic_properties = ('created_at', 'id', 'name', 'owner', 'state',
                          'type_name', 'type_version', 'updated_at',
                          'version', 'visibility', 'description', 'tags',
                          'published_at', 'deleted_at')

    supported_show_levels = ('none', 'basic', 'direct', 'transitive')

    def __init__(self, **kwargs):
        try:
            for prop in self.generic_properties:
                setattr(self, prop, kwargs.pop(prop))
        except KeyError:
            msg = "Invalid parameters were provided"
            raise exc.HTTPBadRequest(msg)
        self.type_specific_properties = {}
        for key, value in kwargs.items():
            try:
                if _is_dependency(value):

                    self.type_specific_properties[key] = ArtifactType(**value)
                elif _is_dependencies_list(value):

                    self.type_specific_properties[key] = [ArtifactType(**elem)
                                                          for elem in value]
                else:
                    self.type_specific_properties[key] = value
            except exc.HTTPBadRequest:
                # if it's not possible to generate artifact object then
                # assign the value as a regular dict.
                self.type_specific_properties[key] = value


def _is_dependency(d):
    if type(d) is dict and d.get('type_name') and d.get('type_version'):
        return True
    return False


def _is_dependencies_list(l):
    if type(l) is list and all(_is_dependency(d) for d in l):
        return True
    return False



#    Copyright (c) 2013 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import urllib

from muranoclient.common import base


class Environment(base.Resource):
    def __repr__(self):
        return "<Environment %s>" % self._info

    def data(self, **kwargs):
        return self.manager.data(self, **kwargs)


class Status(base.Resource):
    def __repr__(self):
        return '<Status %s>' % self._info

    def data(self, **kwargs):
        return self.manager.data(self, **kwargs)


class EnvironmentManager(base.ManagerWithFind):
    resource_class = Environment

    def list(self, all_tenants=False, tenant_id=None):
        params = {'all_tenants': all_tenants}
        if tenant_id:
            params['tenant'] = tenant_id
        path = '/v1/environments?{query}'.format(
            query=urllib.parse.urlencode(params))
        return self._list(path, 'environments')

    def create(self, data):
        return self._create('/v1/environments', data)

    def update(self, environment_id, name):
        return self._update('/v1/environments/{id}'.format(id=environment_id),
                            data={'name': name})

    def delete(self, environment_id, abandon=False):
        path = '/v1/environments/{id}?{query}'.format(
            id=environment_id,
            query=urllib.parse.urlencode({'abandon': abandon}))
        return self._delete(path)

    def get(self, environment_id, session_id=None):
        if session_id:
            headers = {'X-Configuration-Session': session_id}
        else:
            headers = {}
        return self._get("/v1/environments/{id}".format(id=environment_id),
                         headers=headers)

    def last_status(self, environment_id, session_id):
        headers = {'X-Configuration-Session': session_id}
        path = '/v1/environments/{id}/lastStatus'
        path = path.format(id=environment_id)
        status_dict = self._get(path, return_raw=True,
                                response_key='lastStatuses',
                                headers=headers)
        result = {}
        for k, v in status_dict.items():
            if v:
                result[k] = Status(self, v, loaded=True)
        return result

    def get_model(self, environment_id, path, session_id=None):
        headers = {'X-Configuration-Session': session_id}
        url = '/v1/environments/{id}/model/{path}'
        url = url.format(id=environment_id, path=path)
        return self._get(url, return_raw=True, headers=headers)

    def update_model(self, environment_id, data, session_id):
        headers = {'X-Configuration-Session': session_id}
        url = '/v1/environments/{id}/model/'
        url = url.format(id=environment_id)
        return self._update(url, data, return_raw=True, headers=headers,
                            method='PATCH',
                            content_type='application/env-model-json-patch')



# Copyright 2012 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Base utilities to build API operation managers and objects on top of.
"""

import abc
import copy

from muranoclient.apiclient import exceptions


def getid(obj):
    """Get obj's id or object itself if no id

    Abstracts the common pattern of allowing both an object or
    an object's ID (UUID) as a parameter when dealing with relationships.
    """
    try:
        return obj.id
    except AttributeError:
        return obj


class Manager(object):
    """Interacts with type of API

    Managers interact with a particular type of API (servers, flavors,
    images, etc.) and provide CRUD operations for them.
    """
    resource_class = None

    def __init__(self, api):
        self.api = api

    def _list(self, url, response_key=None, obj_class=None,
              data=None, headers=None):

        if headers is None:
            headers = {}
        resp, body = self.api.json_request(url, 'GET', headers=headers)

        if obj_class is None:
            obj_class = self.resource_class

        if response_key:
            if response_key not in body:
                body[response_key] = []
            data = body[response_key]
        else:
            data = body
        return [obj_class(self, res, loaded=True) for res in data if res]

    def _delete(self, url, headers=None):
        if headers is None:
            headers = {}
        self.api.request(url, 'DELETE', headers=headers)

    def _update(self, url, data, response_key=None, return_raw=False,
                headers=None, method='PUT', content_type='application/json'):
        if headers is None:
            headers = {}
        resp, body = self.api.json_request(url, method,
                                           content_type=content_type,
                                           data=data, headers=headers)
        # PUT or PATCH requests may not return a body
        if body:
            if return_raw:
                if response_key:
                    return body[response_key]
                return body
            if response_key:
                return self.resource_class(self, body[response_key])
            return self.resource_class(self, body)

    def _create(self, url, data=None, response_key=None,
                return_raw=False, headers=None):
        if headers is None:
            headers = {}
        if data:
            resp, body = self.api.json_request(url, 'POST',
                                               data=data, headers=headers)
        else:
            resp, body = self.api.json_request(url, 'POST', headers=headers)
        if return_raw:
            if response_key:
                return body[response_key]
            return body
        if response_key:
            return self.resource_class(self, body[response_key])
        return self.resource_class(self, body)

    def _get(self, url, response_key=None, return_raw=False, headers=None):
        if headers is None:
            headers = {}
        resp, body = self.api.json_request(url, 'GET', headers=headers)
        if return_raw:
            if response_key:
                return body[response_key]
            return body
        if response_key:
            return self.resource_class(self, body[response_key])
        return self.resource_class(self, body)


class ManagerWithFind(Manager, metaclass=abc.ABCMeta):
    """Manager with additional `find()`/`findall()` methods."""

    @abc.abstractmethod
    def list(self):
        pass

    def find(self, **kwargs):
        """Find a single item with attributes matching ``**kwargs``.

        This isn't very efficient: it loads the entire list then filters on
        the Python side.
        """
        rl = self.findall(**kwargs)
        num = len(rl)

        if num == 0:
            msg = "No %s matching %s." % (self.resource_class.__name__, kwargs)
            raise exceptions.NotFound(msg)
        elif num > 1:
            raise exceptions.NoUniqueMatch
        else:
            return self.get(rl[0].id)

    def findall(self, **kwargs):
        """Find all items with attributes matching ``**kwargs``.

        This isn't very efficient: it loads the entire list then filters on
        the Python side.
        """
        found = []
        searches = kwargs.items()

        for obj in self.list():
            try:
                if all(getattr(obj, attr) == value
                       for (attr, value) in searches):
                    found.append(obj)
            except AttributeError:
                continue

        return found


class Resource(object):
    """Represents an instance of an object

    A resource represents a particular instance of an object (tenant, user,
    etc). This is pretty much just a bag for attributes.

    :param manager: Manager object
    :param info: dictionary representing resource attributes
    :param loaded: prevent lazy-loading if set to True
    """

    def __init__(self, manager, info, loaded=False):
        self.manager = manager
        self._info = info
        self._add_details(info)
        self._loaded = loaded

    def _add_details(self, info):
        for k, v in info.items():
            setattr(self, k, v)

    def __setstate__(self, d):
        for k, v in d.items():
            setattr(self, k, v)

    def __getattr__(self, k):
        if k not in self.__dict__:
            # NOTE(bcwaldon): disallow lazy-loading if already loaded once
            if not self.is_loaded():
                self.get()
                return self.__getattr__(k)
            raise AttributeError(k)
        else:
            return self.__dict__[k]

    def __repr__(self):
        reprkeys = sorted(k for k in self.__dict__.keys() if k[0] != '_' and
                          k != 'manager')
        info = ", ".join("%s=%s" % (k, getattr(self, k)) for k in reprkeys)
        return "<%s %s>" % (self.__class__.__name__, info)

    def get(self):
        # set_loaded() first ... so if we have to bail, we know we tried.
        self.set_loaded(True)
        if not hasattr(self.manager, 'get'):
            return

        new = self.manager.get(self.id)
        if new:
            self._add_details(new._info)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self._info == other._info

    def __ne__(self, other):
        return not self.__eq__(other)

    def is_loaded(self):
        return self._loaded

    def set_loaded(self, val):
        self._loaded = val

    def to_dict(self):
        return copy.deepcopy(self._info)



# Copyright 2012 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import hashlib
import os
import socket

import keystoneclient.adapter as keystone_adapter
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import encodeutils
import requests
import urllib

from muranoclient.common import exceptions as exc

LOG = logging.getLogger(__name__)
USER_AGENT = 'python-muranoclient'
CHUNKSIZE = 1024 * 64  # 64kB


def get_system_ca_file():
    """Return path to system default CA file."""
    # Standard CA file locations for Debian/Ubuntu, RedHat/Fedora,
    # Suse, FreeBSD/OpenBSD, MacOSX, and the bundled ca
    ca_path = ['/etc/ssl/certs/ca-certificates.crt',
               '/etc/pki/tls/certs/ca-bundle.crt',
               '/etc/ssl/ca-bundle.pem',
               '/etc/ssl/cert.pem',
               '/System/Library/OpenSSL/certs/cacert.pem',
               requests.certs.where()]
    for ca in ca_path:
        LOG.debug("Looking for ca file %s", ca)
        if os.path.exists(ca):
            LOG.debug("Using ca file %s", ca)
            return ca
    LOG.warning("System ca file could not be found.")


class HTTPClient(object):

    def __init__(self, endpoint, **kwargs):
        self.endpoint = endpoint
        self.auth_url = kwargs.get('auth_url')
        self.auth_token = kwargs.get('token')
        self.username = kwargs.get('username')
        self.password = kwargs.get('password')
        self.region_name = kwargs.get('region_name')
        self.include_pass = kwargs.get('include_pass')
        self.endpoint_url = endpoint

        self.cert_file = kwargs.get('cert_file')
        self.key_file = kwargs.get('key_file')
        self.timeout = kwargs.get('timeout')

        self.ssl_connection_params = {
            'cacert': kwargs.get('cacert'),
            'cert_file': kwargs.get('cert_file'),
            'key_file': kwargs.get('key_file'),
            'insecure': kwargs.get('insecure'),
        }

        self.verify_cert = None
        if urllib.parse.urlparse(endpoint).scheme == "https":
            if kwargs.get('insecure'):
                self.verify_cert = False
            else:
                self.verify_cert = kwargs.get('cacert', get_system_ca_file())

    def _safe_header(self, name, value):
        if name in ['X-Auth-Token', 'X-Subject-Token']:
            # because in python3 byte string handling is ... ug
            v = value.encode('utf-8')
            h = hashlib.sha1(v)
            d = h.hexdigest()
            return encodeutils.safe_decode(name), "{SHA1}%s" % d
        else:
            return (encodeutils.safe_decode(name),
                    encodeutils.safe_decode(value))

    def log_curl_request(self, url, method, kwargs):
        curl = ['curl -i -X %s' % method]

        for (key, value) in kwargs['headers'].items():
            header = '-H \'%s: %s\'' % self._safe_header(key, value)
            curl.append(header)

        conn_params_fmt = [
            ('key_file', '--key %s'),
            ('cert_file', '--cert %s'),
            ('cacert', '--cacert %s'),
        ]
        for (key, fmt) in conn_params_fmt:
            value = self.ssl_connection_params.get(key)
            if value:
                curl.append(fmt % value)

        if self.ssl_connection_params.get('insecure'):
            curl.append('-k')

        if 'data' in kwargs:
            curl.append('-d \'%s\'' % kwargs['data'])

        curl.append('%s%s' % (self.endpoint, url))
        LOG.debug(' '.join(curl))

    @staticmethod
    def log_http_response(resp):
        status = (resp.raw.version / 10.0, resp.status_code, resp.reason)
        dump = ['\nHTTP/%.1f %s %s' % status]
        dump.extend(['%s: %s' % (k, v) for k, v in resp.headers.items()])
        dump.append('')
        if resp.content:
            content = resp.content
            if isinstance(content, bytes):
                try:
                    content = encodeutils.safe_decode(resp.content)
                except UnicodeDecodeError:
                    pass
                else:
                    dump.extend([content, ''])
        LOG.debug('\n'.join(dump))

    def request(self, url, method, log=True, **kwargs):
        """Send an http request with the specified characteristics.

        Wrapper around requests.request to handle tasks such
        as setting headers and error handling.
        """
        _set_data(kwargs)

        # Copy the kwargs so we can reuse the original in case of redirects
        kwargs['headers'] = copy.deepcopy(kwargs.get('headers', {}))
        kwargs['headers'].setdefault('User-Agent', USER_AGENT)
        if self.auth_token:
            kwargs['headers'].setdefault('X-Auth-Token', self.auth_token)
        else:
            kwargs['headers'].update(self.credentials_headers())
        if self.auth_url:
            kwargs['headers'].setdefault('X-Auth-Url', self.auth_url)
        if self.region_name:
            kwargs['headers'].setdefault('X-Region-Name', self.region_name)

        self.log_curl_request(url, method, kwargs)

        if self.cert_file and self.key_file:
            kwargs['cert'] = (self.cert_file, self.key_file)

        if self.verify_cert is not None:
            kwargs['verify'] = self.verify_cert

        if self.timeout is not None:
            kwargs['timeout'] = float(self.timeout)

        # Allow the option not to follow redirects
        follow_redirects = kwargs.pop('follow_redirects', True)

        # Since requests does not follow the RFC when doing redirection to sent
        # back the same method on a redirect we are simply bypassing it.  For
        # example if we do a DELETE/POST/PUT on a URL and we get a 302 RFC says
        # that we should follow that URL with the same method as before,
        # requests doesn't follow that and send a GET instead for the method.
        # Hopefully this could be fixed as they say in a comment in a future
        # point version i.e.: 3.x
        # See issue: https://github.com/kennethreitz/requests/issues/1704
        allow_redirects = False

        try:
            resp = requests.request(
                method,
                self.endpoint_url + url,
                allow_redirects=allow_redirects,
                **kwargs)
        except socket.gaierror as e:
            message = ("Error finding address for %(url)s: %(e)s" %
                       {'url': self.endpoint_url + url, 'e': e})
            raise exc.InvalidEndpoint(message=message)
        except (socket.error,
                socket.timeout,
                requests.exceptions.ConnectionError) as e:
            endpoint = self.endpoint
            message = ("Error communicating with %(endpoint)s %(e)s" %
                       {'endpoint': endpoint, 'e': e})
            raise exc.CommunicationError(message=message)

        if log:
            self.log_http_response(resp)

        if 'X-Auth-Key' not in kwargs['headers'] and \
                (resp.status_code == 401 or
                 (resp.status_code == 500 and
                  "(HTTP 401)" in resp.content)):
            raise exc.HTTPUnauthorized("Authentication failed. Please try"
                                       " again.\n%s"
                                       % resp.content)
        elif 400 <= resp.status_code < 600:
            raise exc.from_response(resp)
        elif resp.status_code in (301, 302, 305):
            # Redirected. Reissue the request to the new location,
            # unless caller specified follow_redirects=False
            if follow_redirects:
                location = resp.headers.get('location')
                path = self.strip_endpoint(location)
                resp = self.request(path, method, **kwargs)
        elif resp.status_code == 300:
            raise exc.from_response(resp)

        return resp

    def strip_endpoint(self, location):
        if location is None:
            message = "Location not returned with 302"
            raise exc.InvalidEndpoint(message=message)
        elif location.startswith(self.endpoint):
            return location[len(self.endpoint):]
        else:
            message = "Prohibited endpoint redirect %s" % location
            raise exc.InvalidEndpoint(message=message)

    def credentials_headers(self):
        creds = {}
        if self.username:
            creds['X-Auth-User'] = self.username
        if self.password:
            creds['X-Auth-Key'] = self.password
        return creds

    def json_request(self, url, method, content_type='application/json',
                     **kwargs):

        kwargs.setdefault('headers', {})
        kwargs['headers'].setdefault('Content-Type', content_type)
        # Don't set Accept because we aren't always dealing in JSON

        _set_data(kwargs)
        if 'data' in kwargs:
            kwargs['data'] = jsonutils.dumps(kwargs['data'])

        resp = self.request(url, method, **kwargs)
        body = resp.content

        if body and 'application/json' in resp.headers['content-type']:
            try:
                body = resp.json()
            except ValueError:
                LOG.error('Could not decode response body as JSON')
        else:
            body = None

        return resp, body

    def json_patch_request(self, url, method='PATCH', **kwargs):
        content_type = 'application/murano-packages-json-patch'
        return self.json_request(
            url, method, content_type=content_type, **kwargs)

    def head(self, url, **kwargs):
        return self.json_request(url, "HEAD", **kwargs)

    def get(self, url, **kwargs):
        return self.json_request(url, "GET", **kwargs)

    def post(self, url, **kwargs):
        return self.json_request(url, "POST", **kwargs)

    def put(self, url, **kwargs):
        return self.json_request(url, "PUT", **kwargs)

    def delete(self, url, **kwargs):
        return self.request(url, "DELETE", **kwargs)

    def patch(self, url, **kwargs):
        return self.json_request(url, "PATCH", **kwargs)


class SessionClient(keystone_adapter.Adapter):
    """Murano specific keystoneclient Adapter.

    Murano can't use keystoneclient LegacyJsonAdapter, because murano has the
    check for right content-type for "update" operation which is
    'application/murano-packages-json-patch'. So, we need to create our own
    adapter.
    """

    def request(self, url, method, **kwargs):
        raise_exc = kwargs.pop('raise_exc', True)
        _set_data(kwargs)
        resp = super(SessionClient, self).request(url,
                                                  method,
                                                  raise_exc=False,
                                                  **kwargs)

        if raise_exc and resp.status_code >= 400:
            LOG.trace("Error communicating with {url}: {exc}"
                      .format(url=url, exc=exc.from_response(resp)))
            raise exc.from_response(resp)

        return resp

    def json_request(self, url, method, **kwargs):
        headers = kwargs.setdefault('headers', {})
        headers['Content-Type'] = kwargs.pop('content_type',
                                             'application/json')

        _set_data(kwargs)
        if 'data' in kwargs:
            kwargs['data'] = jsonutils.dumps(kwargs['data'])
            # NOTE(starodubcevna): We need to prove that json field is empty,
            # or it will be modified by keystone adapter.
            kwargs['json'] = None

        resp = self.request(url, method, **kwargs)
        body = resp.text
        if body:
            try:
                body = jsonutils.loads(body)
            except ValueError:
                pass

        return resp, body

    def json_patch_request(self, url, method='PATCH', **kwargs):
        content_type = 'application/murano-packages-json-patch'
        return self.json_request(
            url, method, content_type=content_type, **kwargs)


def _construct_http_client(*args, **kwargs):
    session = kwargs.pop('session', None)
    auth = kwargs.pop('auth', None)
    endpoint = next(iter(args), None)

    if session:
        service_type = kwargs.pop('service_type', None)
        endpoint_type = kwargs.pop('endpoint_type', None)
        region_name = kwargs.pop('region_name', None)
        service_name = kwargs.pop('service_name', None)
        parameters = {
            'endpoint_override': endpoint,
            'session': session,
            'auth': auth,
            'interface': endpoint_type,
            'service_type': service_type,
            'region_name': region_name,
            'service_name': service_name,
            'user_agent': 'python-muranoclient',
        }
        parameters.update(kwargs)
        return SessionClient(**parameters)
    else:
        return HTTPClient(*args, **kwargs)


def _set_data(kwargs):
    if 'body' in kwargs:
        if 'data' in kwargs:
            raise ValueError("Can't provide both 'data' and "
                             "'body' to a request")
        LOG.warning("Use of 'body' is deprecated; use 'data' instead")
        kwargs['data'] = kwargs.pop('body')



#    Copyright (c) 2013 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import functools
import itertools
import json
import os
import shutil
import sys
import tempfile
import zipfile

import jsonpatch
from oslo_utils import strutils
from oslo_utils import uuidutils
import urllib

from muranoclient.apiclient import exceptions
from muranoclient.common import exceptions as common_exceptions
from muranoclient.common import utils
from muranoclient.v1.package_creator import hot_package
from muranoclient.v1.package_creator import mpl_package

_bool_from_str_strict = functools.partial(
    strutils.bool_from_string, strict=True)


@utils.arg('--all-tenants', action='store_true', default=False,
           help='Allows to list environments from all tenants'
                ' (admin only).')
@utils.arg('--tenant', metavar="<TENANT_ID>", default=None,
           help='Allows to list environments for a given tenant'
                ' (admin only).')
def do_environment_list(mc, args=None):
    """List the environments."""
    if args is None:
        args = {}
    all_tenants = getattr(args, 'all_tenants', False)
    tenant = getattr(args, 'tenant', None)
    environments = mc.environments.list(all_tenants, tenant)
    _print_environment_list(environments)


def _print_environment_list(environments):
    field_labels = ['ID', 'Name', 'Status', 'Created', 'Updated']
    fields = ['id', 'name', 'status', 'created', 'updated']
    utils.print_list(environments, fields, field_labels, sortby=0)


def _generate_join_existing_net(net, subnet):
    res = {
        'defaultNetworks': {
            'environment': {
                '?': {
                    'id': uuidutils.generate_uuid(dashed=False),
                    'type': 'io.murano.resources.ExistingNeutronNetwork'
                },
            },
            'flat': None
        }
    }
    if net:
        res['defaultNetworks']['environment']['internalNetworkName'] = net
    if subnet:
        res['defaultNetworks']['environment']['internalSubnetworkName'] = \
            subnet
    return res


@utils.arg("--join-net-id", metavar="<NET_ID>",
           help="Network id to join.",)
@utils.arg("--join-subnet-id", metavar="<SUBNET_ID>",
           help="Subnetwork id to join.",)
@utils.arg("--region", metavar="<REGION_NAME>",
           help="Name of the target OpenStack region.",)
@utils.arg("name", metavar="<ENVIRONMENT_NAME>",
           help="Environment name.")
def do_environment_create(mc, args):
    """Create an environment."""
    body = {"name": args.name, "region": args.region}
    if args.join_net_id or args.join_subnet_id:
        body.update(_generate_join_existing_net(
            args.join_net_id, args.join_subnet_id))
    environment = mc.environments.create(body)
    _print_environment_list([environment])


@utils.arg("id", metavar="<NAME or ID>",
           nargs="+", help="Id or name of environment(s) to delete.")
@utils.arg('--abandon', action='store_true', default=False,
           help='If set will abandon environment without deleting any'
                ' of its resources.')
def do_environment_delete(mc, args):
    """Delete an environment."""
    abandon = getattr(args, 'abandon', False)
    failure_count = 0
    for environment_id in args.id:
        try:
            environment = utils.find_resource(mc.environments, environment_id)
            mc.environments.delete(environment.id, abandon)
        except exceptions.NotFound:
            failure_count += 1
            print("Failed to delete '{0}'; environment not found".
                  format(environment_id))
    if failure_count == len(args.id):
        raise exceptions.CommandError("Unable to find and delete any of the "
                                      "specified environments.")
    do_environment_list(mc)


@utils.arg("id", metavar="<NAME or ID>",
           help="Environment ID or name.")
@utils.arg("name", metavar="<ENVIRONMENT_NAME>",
           help="A name to which the environment will be renamed.")
def do_environment_rename(mc, args):
    """Rename an environment."""
    try:
        environment = utils.find_resource(mc.environments, args.id)
        environment = mc.environments.update(environment.id, args.name)
    except exceptions.NotFound:
        raise exceptions.CommandError("Environment %s not found" % args.id)
    else:
        _print_environment_list([environment])


@utils.arg("id", metavar="<NAME or ID>",
           help="Environment ID or name.")
@utils.arg("--session-id", metavar="<SESSION_ID>", default='',
           help="Id of a config session.")
@utils.arg("--only-apps", action='store_true',
           help="Only print apps of the environment (useful for automation).")
def do_environment_show(mc, args):
    """Display environment details."""
    try:
        environment = utils.find_resource(
            mc.environments, args.id, session_id=args.session_id)
    except exceptions.NotFound:
        raise exceptions.CommandError("Environment %s not found" % args.id)
    else:
        if getattr(args, 'only_apps', False):
            print(utils.json_formatter(environment.services))
        else:
            formatters = {
                "id": utils.text_wrap_formatter,
                "created": utils.text_wrap_formatter,
                "name": utils.text_wrap_formatter,
                "tenant_id": utils.text_wrap_formatter,
                "services": utils.json_formatter,

            }
            utils.print_dict(environment.to_dict(), formatters=formatters)


@utils.arg("id", metavar="<ID>",
           help="ID of Environment to deploy.")
@utils.arg("--session-id", metavar="<SESSION>",
           required=True,
           help="ID of configuration session to deploy.")
def do_environment_deploy(mc, args):
    """Start deployment of a murano environment session."""
    mc.sessions.deploy(args.id, args.session_id)
    do_environment_show(mc, args)


@utils.arg("id", help="ID of Environment to call action against.")
@utils.arg("--action-id", metavar="<ACTION>",
           required=True,
           help="ID of action to run.")
@utils.arg("--arguments", metavar='<KEY=VALUE>', nargs='*',
           help="Action arguments.")
def do_environment_action_call(mc, args):
    """Call action `ACTION` in environment `ID`.

    Returns id of an asynchronous task, that executes the action.
    Actions can only be called on a `deployed` environment.
    To view actions available in a given environment use `environment-show`
    command.
    """
    arguments = {}
    for argument in args.arguments or []:
        if '=' not in argument:
            raise exceptions.CommandError(
                "Argument should be in form of KEY=VALUE. Found: {0}".format(
                    argument))
        k, v = argument.split('=', 1)
        try:
            v = json.loads(v)
        except ValueError:
            # treat value as a string if it doesn't load as json
            pass
        arguments[k] = v
    task_id = mc.actions.call(
        args.id, args.action_id, arguments=arguments)
    print("Created task, id: {0}".format(task_id))


@utils.arg("id", metavar="<ID>",
           help="ID of Environment where task is being executed.")
@utils.arg("--task-id", metavar="<TASK>",
           required=True,
           help="ID of action to run.")
def do_environment_action_get_result(mc, args):
    """Get result of `TASK` in environment `ID`."""
    result = mc.actions.get_result(args.id, args.task_id)
    print("Task id result: {0}".format(result))


@utils.arg("class_name", metavar='<CLASS>',
           help="FQN of the class with static method")
@utils.arg("method_name", metavar='<METHOD>', help="Static method to run")
@utils.arg("--arguments", metavar='<KEY=VALUE>', nargs='*',
           help="Method arguments. No arguments by default")
@utils.arg("--package-name", metavar='<PACKAGE>', default='',
           help='Optional FQN of the package to look for the class in')
@utils.arg("--class-version", default='',
           help='Optional version of the class, otherwise version =0 is '
                'used ')
def do_static_action_call(mc, args):
    """Call static method `METHOD` of the class `CLASS` with `ARGUMENTS`.

    Returns the result of the method execution.
    `PACKAGE` and `CLASS_VERSION` can be specified optionally to find class in
    a particular package and to look for the specific version of a class
    respectively.
    """
    arguments = {}
    for argument in args.arguments or []:
        if '=' not in argument:
            raise exceptions.CommandError(
                "Argument should be in form of KEY=VALUE. Found: {0}".format(
                    argument))
        key, value = argument.split('=', 1)
        try:
            value = json.loads(value)
        except ValueError:
            # treat value as a string if it doesn't load as json
            pass
        arguments[key] = value

    request_body = {
        "className": args.class_name,
        "methodName": args.method_name,
        "packageName": args.package_name or None,
        "classVersion": args.class_version or '=0',
        "parameters": arguments
    }

    print("Waiting for result...")
    try:
        result = mc.static_actions.call(request_body).get_result()
        print("Static action result: {0}".format(result))
    except Exception as e:
        print(str(e))


@utils.arg("id", metavar="<ID>", help="ID of Environment to add session to.")
def do_environment_session_create(mc, args):
    """Creates a new configuration session for environment ID."""
    environment_id = args.id
    session_id = mc.sessions.configure(environment_id).id
    print("Created new session:")
    formatters = {"id": utils.text_wrap_formatter}
    utils.print_dict({"id": session_id}, formatters=formatters)


@utils.arg("id", metavar="<ID>", help="ID of Environment to edit.")
@utils.arg("filename", metavar="FILE", nargs="?",
           help="File to read jsonpatch from (defaults to stdin).")
@utils.arg("--session-id", metavar="<SESSION_ID>",
           required=True,
           help="Id of a config session.")
def do_environment_apps_edit(mc, args):
    """Edit environment's object model.

    `FILE` is path to a file, that contains jsonpatch, that describes changes
    to be made to environment's object-model.

    [
        { "op": "add", "path": "/-",
           "value": { ... your-app object model here ... }
        },
        { "op": "replace", "path": "/0/?/name",
          "value": "new_name"
        },
    ]

    NOTE: Values '===id1===', '===id2===', etc. in the resulting object-model
    will be substituted with uuids.

    For more info on jsonpatch see RFC 6902
    """

    jp_obj = None
    if not args.filename:
        jp_obj = json.load(sys.stdin)
    else:
        with open(args.filename) as fpatch:
            jp_obj = json.load(fpatch)

    jpatch = jsonpatch.JsonPatch(jp_obj)

    environment_id = args.id
    session_id = args.session_id
    environment = mc.environments.get(environment_id, session_id)

    object_model = jpatch.apply(environment.services)
    utils.traverse_and_replace(object_model)

    mc.services.put(
        environment_id,
        path='/',
        data=jpatch.apply(environment.services),
        session_id=session_id)


@utils.arg("id", metavar="<ID>", help="ID of Environment to show.")
@utils.arg("--path", metavar="<PATH>", default='/',
           help="Path to Environment model section. Defaults to '/'.")
@utils.arg("--session-id", metavar="<SESSION_ID>",
           help="Id of a config session.")
def do_environment_model_show(mc, args):
    """Display an environment's object model."""
    session_id = args.session_id or None
    path = urllib.parse.quote(args.path)
    env_model = mc.environments.get_model(args.id, path, session_id)
    print(utils.json_formatter(env_model))


@utils.arg("id", metavar="<ID>", help="ID of Environment to edit.")
@utils.arg("filename", metavar="<FILE>", nargs="?",
           help="File to read JSON-patch from (defaults to stdin).")
@utils.arg("--session-id", metavar="<SESSION_ID>", required=True,
           help="Id of a config session.")
def do_environment_model_edit(mc, args):
    """Edit an environment's object model."""
    jp_obj = None
    if not args.filename:
        jp_obj = json.load(sys.stdin)
    else:
        with open(args.filename) as fpatch:
            jp_obj = json.load(fpatch)

    if not isinstance(jp_obj, list):
        raise exceptions.CommandError('JSON-patch must be a list of changes')
    for change in jp_obj:
        if 'op' not in change or 'path' not in change:
            raise exceptions.CommandError('Every change in JSON-patch must '
                                          'contain "op" and "path" keys')
        op = change['op']
        if op not in ['add', 'replace', 'remove']:
            raise exceptions.CommandError('The value of "op" item must be '
                                          '"add", "replace" or "remove", '
                                          'got {0}'.format(op))
        if op != 'remove' and 'value' not in change:
            raise exceptions.CommandError('"add" or "replace" change in '
                                          'JSON-patch must contain "value" '
                                          'key')
    session_id = args.session_id
    new_model = mc.environments.update_model(args.id, jp_obj, session_id)
    print(utils.json_formatter(new_model))


def do_env_template_list(mc, args=None):
    """List the environments templates."""
    if args is None:
        args = {}
    env_templates = mc.env_templates.list()
    _print_env_template_list(env_templates)


def _print_env_template_list(env_templates):
    field_labels = ['ID', 'Name', 'Created', 'Updated', 'Is public']
    fields = ['id', 'name', 'created', 'updated', 'is_public']
    utils.print_list(env_templates, fields, field_labels, sortby=0)


@utils.arg("name", metavar="<ENV_TEMPLATE_NAME>",
           help="Environment template name.")
@utils.arg("--is-public", action='store_true', default=False,
           help='Make the template available for users from other tenants.')
def do_env_template_create(mc, args):
    """Create an environment template."""
    env_template = mc.env_templates.create(
        {"name": args.name, "is_public": args.is_public})
    _print_env_template_list([env_template])


@utils.arg("id", metavar="<ID>",
           help="Environment template ID.")
@utils.arg("name", metavar="<ENV_NAME>",
           help="New environment name.")
@utils.arg("--region", metavar="<REGION_NAME>",
           help="Name of the target OpenStack region.",)
def do_env_template_create_env(mc, args):
    """Create a new environment from template."""
    try:
        data = {}
        data["name"] = args.name
        if args.region:
            data["region"] = args.region
        template = mc.env_templates.create_env(args.id, data)
    except common_exceptions.HTTPNotFound:
        raise exceptions.CommandError("Environment template %s not found"
                                      % args.id)
    else:
        formatters = {
            "environment_id": utils.text_wrap_formatter,
            "session_id": utils.text_wrap_formatter
        }
        utils.print_dict(template.to_dict(), formatters=formatters)


@utils.arg("id", metavar="<ID>",
           help="Environment template ID.")
def do_env_template_show(mc, args):
    """Display environment template details."""
    try:
        env_template = mc.env_templates.get(args.id)
    except common_exceptions.HTTPNotFound:
        raise exceptions.CommandError("Environment template %s not found"
                                      % args.id)
    else:
        formatters = {
            "id": utils.text_wrap_formatter,
            "created": utils.text_wrap_formatter,
            "name": utils.text_wrap_formatter,
            "tenant_id": utils.text_wrap_formatter,
            "services": utils.json_formatter,

        }
        utils.print_dict(env_template.to_dict(), formatters=formatters)


@utils.arg("id", metavar="<ENV_TEMPLATE_ID>",
           help="Environment template ID.")
@utils.arg('app_template_file', metavar='<FILE>',
           help='Path to the template.')
def do_env_template_add_app(mc, args):
    """Add application to the environment template."""
    with open(args.app_template_file, "r") as app_file:
        app_templates = json.load(app_file)
    if not isinstance(app_templates, list):
        app_templates = [app_templates]
    for app_template in app_templates:
        mc.env_templates.create_app(args.id, app_template)
    do_env_template_show(mc, args)


@utils.arg("id", metavar="<ENV_TEMPLATE_ID>",
           help="Environment template ID.")
@utils.arg("app_id", metavar="<ENV_TEMPLATE_APP_ID>",
           help="Application ID.")
def do_env_template_del_app(mc, args):
    """Delete application from the environment template."""
    mc.env_templates.delete_app(args.id, args.app_id)
    do_env_template_show(mc, args)


@utils.arg("id", metavar="<ID>",
           help="Environment template ID.")
@utils.arg("name", metavar="<ENV_TEMPLATE_NAME>",
           help="Environment template name.")
def do_env_template_update(mc, args):
    """Update an environment template."""
    try:
        env_template = mc.env_templates.update(args.id, args.name)
    except common_exceptions.HTTPNotFound:
        raise exceptions.CommandError("Environment template %s not found"
                                      % args.id)
    _print_env_template_list([env_template])


@utils.arg("id", metavar="<ID>",
           nargs="+", help="ID of environment(s) template to delete.")
def do_env_template_delete(mc, args):
    """Delete an environment template."""
    failure_count = 0
    for env_template_id in args.id:
        try:
            mc.env_templates.delete(env_template_id)
        except common_exceptions.HTTPNotFound:
            failure_count += 1
            mns = "Failed to delete '{0}'; environment template not found".\
                format(env_template_id)

    if failure_count == len(args.id):
        raise exceptions.CommandError(mns)
    do_env_template_list(mc)


@utils.arg("id", metavar="<ID>",
           help="Environment template ID.")
@utils.arg("name", metavar="<ENV_TEMPLATE_NAME>",
           help="New environment template name.")
def do_env_template_clone(mc, args):
    """Create a new template, cloned from template."""
    try:
        env_template = mc.env_templates.clone(args.id, args.name)
    except common_exceptions.HTTPNotFound:
        raise exceptions.CommandError("Environment template %s not found"
                                      % args.id)
    else:
        formatters = {
            "id": utils.text_wrap_formatter,
            "created": utils.text_wrap_formatter,
            "updated": utils.text_wrap_formatter,
            "version": utils.text_wrap_formatter,
            "name": utils.text_wrap_formatter,
            "tenant_id": utils.text_wrap_formatter,
            "is_public": utils.text_wrap_formatter,
            "services": utils.json_formatter,

        }
        utils.print_dict(env_template.to_dict(), formatters=formatters)


@utils.arg("id", metavar="<ID>", nargs='?', default=None,
           help="Environment ID for which to list deployments.")
@utils.arg('--all-environments', action='store_true', default=False,
           help="Lists all deployments for all environments in user's tenant.")
def do_deployment_list(mc, args):
    """List deployments for an environment or multiple environments."""
    all_environments = getattr(args, 'all_environments', False)
    env_id = getattr(args, 'id', None)
    if env_id and all_environments:
        raise exceptions.CommandError(
            'Environment ID and all-environments flag cannot both be set.')
    elif not env_id and not all_environments:
        raise exceptions.CommandError(
            'Either environment ID or all-environments flag must be set.')
    try:
        if all_environments:
            deployments = mc.deployments.list(None, all_environments)
        else:
            environment = utils.find_resource(mc.environments, env_id)
            deployments = mc.deployments.list(environment.id)
    except exceptions.NotFound:
        if env_id:
            msg = "Environment %s not found" % env_id
        else:
            msg = "Environments not found"
        raise exceptions.CommandError(msg)
    else:
        field_labels = ["ID", "State", "Created", "Updated", "Finished"]
        fields = ["id", "state", "created", "updated", "finished"]
        utils.print_list(deployments, fields, field_labels, sortby=0)


@utils.arg("--limit", type=int, default=0,
           help='Show limited number of packages')
@utils.arg("--marker", default='',
           help='Show packages starting from package with id excluding it')
@utils.arg("--include-disabled", default=False, action="store_true")
@utils.arg("--owned", default=False, action="store_true")
@utils.arg('--search', metavar='<SEARCH_KEYS>',
           dest='search', required=False,
           help='Show packages, that match search keys fuzzily')
@utils.arg('--name', metavar='<PACKAGE_NAME>',
           dest='name', required=False,
           help='Show packages, whose name match parameter exactly')
@utils.arg('--fqn', metavar="<PACKAGE_FULLY_QUALIFIED_NAME>",
           dest='fqn', required=False,
           help='Show packages, '
                'whose fully qualified name match parameter exactly')
@utils.arg('--type', metavar='<PACKAGE_TYPE>',
           dest='type', required=False,
           help='Show packages, whose type match parameter exactly')
@utils.arg('--category', metavar='<PACKAGE_CATEGORY>',
           dest='category', required=False,
           help='Show packages, whose categories include parameter')
@utils.arg('--class_name', metavar='<PACKAGE_CLASS_NAME>',
           dest='class_name', required=False,
           help='Show packages, whose class name match parameter exactly')
@utils.arg('--tag', metavar='<PACKAGE_TAG>',
           dest='tag', required=False,
           help='Show packages, whose tags include parameter')
def do_package_list(mc, args=None):
    """List available packages."""
    filter_args = {
        "include_disabled": getattr(args, 'include_disabled', False),
        "owned": getattr(args, 'owned', False),
    }
    if args:
        if args.limit < 0:
            raise exceptions.CommandError(
                '--limit parameter must be non-negative')
        if args.limit != 0:
            filter_args['limit'] = args.limit
        if args.marker:
            filter_args['marker'] = args.marker
        if args.search:
            filter_args['search'] = args.search
        if args.name:
            filter_args['name'] = args.name
        if args.fqn:
            filter_args['fqn'] = args.fqn
        if args.type:
            filter_args['type'] = args.type
        if args.category:
            filter_args['category'] = args.category
        if args.class_name:
            filter_args['class_name'] = args.class_name
        if args.tag:
            filter_args['tag'] = args.tag

    packages = mc.packages.filter(**filter_args)
    if not args or args.limit == 0:
        _print_package_list(packages)
    else:
        _print_package_list(itertools.islice(packages, args.limit))


def _print_package_list(packages):
    field_labels = ["ID", "Name", "FQN", "Author", "Active",
                    "Is Public", "Type", "Version"]
    fields = ["id", "name", "fully_qualified_name", "author",
              "enabled", "is_public", "type", "version"]
    utils.print_list(packages, fields, field_labels, sortby=0)


@utils.arg("id", metavar="<ID>",
           help="Package ID to download.")
@utils.arg("filename", metavar="file", nargs="?",
           help="Filename to save package to. If it is not specified and "
                "there is no stdout redirection the package won't be saved.")
def do_package_download(mc, args):
    """Download a package to a filename or stdout."""

    def download_to_fh(package_id, fh):
        fh.write(mc.packages.download(package_id))

    try:
        if args.filename:
            with open(args.filename, 'wb') as fh:
                download_to_fh(args.id, fh)
                print("Package downloaded to %s" % args.filename)
        elif not sys.stdout.isatty():
            download_to_fh(args.id, sys.stdout)
        else:
            msg = ('No stdout redirection or local file specified for '
                   'downloaded package. Please specify a local file to save '
                   'downloaded package or redirect output to another source.')
            raise exceptions.CommandError(msg)
    except common_exceptions.HTTPNotFound:
        raise exceptions.CommandError("Package %s not found" % args.id)


@utils.arg("id", metavar="<ID>",
           help="Package ID to show.")
def do_package_show(mc, args):
    """Display details for a package."""
    try:
        package = mc.packages.get(args.id)
    except common_exceptions.HTTPNotFound:
        raise exceptions.CommandError("Package %s not found" % args.id)
    else:
        to_display = dict(
            id=package.id,
            type=package.type,
            owner_id=package.owner_id,
            name=package.name,
            fully_qualified_name=package.fully_qualified_name,
            is_public=package.is_public,
            enabled=package.enabled,
            class_definitions=", ".join(package.class_definitions),
            categories=", ".join(package.categories),
            tags=", ".join(package.tags),
            description=package.description
        )
        formatters = {
            'class_definitions': utils.text_wrap_formatter,
            'categories': utils.text_wrap_formatter,
            'tags': utils.text_wrap_formatter,
            'description': utils.text_wrap_formatter,
        }
        utils.print_dict(to_display, formatters)


@utils.arg("id", metavar="<ID>",
           nargs='+', help="Package ID to delete.")
def do_package_delete(mc, args):
    """Delete a package."""
    failure_count = 0
    for package_id in args.id:
        try:
            mc.packages.delete(package_id)
            print("Deleted package '{0}'".format(package_id))
        except exceptions.NotFound:
            failure_count += 1
            print("Failed to delete '{0}'; package not found".
                  format(package_id))

    if failure_count == len(args.id):
        raise exceptions.CommandError("Unable to find and delete any of the "
                                      "specified packages.")
    else:
        do_package_list(mc)


def _handle_package_exists(mc, data, package, exists_action):
    name = package.manifest['FullName']
    version = package.manifest.get('Version', '0')
    while True:
        print("Importing package {0}".format(name))
        try:
            return mc.packages.create(data, {name: package.file()})
        except common_exceptions.HTTPConflict:
            print("Importing package {0} failed. Package with the same"
                  " name/classes is already registered.".format(name))
            allowed_results = ['s', 'u', 'a']
            res = exists_action
            if not res:
                while True:
                    print("What do you want to do? (s)kip, (u)pdate, (a)bort")
                    res = input()
                    if res in allowed_results:
                        break
            if res == 's':
                print("Skipping.")
                return None
            elif res == 'a':
                print("Exiting.")
                sys.exit()
            elif res == 'u':
                pkgs = list(mc.packages.filter(fqn=name, version=version,
                                               owned=True))
                if not pkgs:
                    msg = (
                        "Got a conflict response, but could not find the "
                        "package '{0}' in the current tenant.\nThis probably "
                        "means the conflicting package is in another tenant.\n"
                        "Please delete it manually."
                    ).format(name)
                    raise exceptions.CommandError(msg)
                elif len(pkgs) > 1:
                    msg = (
                        "Got {0} packages with name '{1}'.\nI do not trust "
                        "myself, please delete the package manually."
                    ).format(len(pkgs), name)
                    raise exceptions.CommandError(msg)
                print("Deleting package {0}({1})".format(name, pkgs[0].id))
                mc.packages.delete(pkgs[0].id)
                continue


@utils.arg('filename', metavar='<FILE>',
           nargs='+',
           help='URL of the murano zip package, FQPN, path to zip package'
                ' or path to directory with package.')
@utils.arg('-c', '--categories', metavar='<CATEGORY>', nargs='*',
           help='Category list to attach.')
@utils.arg('--is-public', action='store_true', default=False,
           help='Make the package available for users from other tenants.')
@utils.arg('--package-version', default='',
           help='Version of the package to use from repository '
                '(ignored when importing with multiple packages).')
@utils.arg('--exists-action', default='', choices=['a', 's', 'u'],
           help='Default action when a package already exists: '
                '(s)kip, (u)pdate, (a)bort.')
@utils.arg('--dep-exists-action', default='', choices=['a', 's', 'u'],
           help='Default action when a dependency package already exists: '
                '(s)kip, (u)pdate, (a)bort.')
def do_package_import(mc, args):
    """Import a package.

    `FILE` can be either a path to a zip file, url or a FQPN.
    You can use `--` to separate `FILE`s from other arguments.

    Categories have to be separated with a space and have to be already
    present in murano.
    """
    data = {"is_public": args.is_public}
    exception_occurred = False
    version = args.package_version
    if version and len(args.filename) >= 2:
        print("Requested to import more than one package, "
              "ignoring version.")
        version = ''

    if args.categories:
        data["categories"] = args.categories

    total_reqs = collections.OrderedDict()
    main_packages_names = []
    for filename in args.filename:
        if os.path.isfile(filename) or os.path.isdir(filename):
            _file = filename
        else:
            print("Package file '{0}' does not exist, attempting to download"
                  "".format(filename))
            _file = utils.to_url(
                filename,
                version=version,
                base_url=args.murano_repo_url,
                extension='.zip',
                path='apps/',
            )
        try:
            package = utils.Package.from_file(_file)
        except Exception as e:
            print("Failed to create package for '{0}', reason: {1}".format(
                filename, e))
            exception_occurred = True
            continue
        total_reqs.update(package.requirements(base_url=args.murano_repo_url))
        main_packages_names.append(package.manifest['FullName'])

    imported_list = []

    dep_exists_action = args.dep_exists_action
    if dep_exists_action == '':
        dep_exists_action = args.exists_action

    for name, package in total_reqs.items():
        image_specs = package.images()
        if image_specs:
            print("Inspecting required images")
            try:
                imgs = utils.ensure_images(
                    glance_client=mc.glance_client,
                    image_specs=image_specs,
                    base_url=args.murano_repo_url,
                    is_package_public=args.is_public)
                for img in imgs:
                    print("Added {0}, {1} image".format(
                        img['name'], img['id']))
            except Exception as e:
                print("Error {0} occurred while installing "
                      "images for {1}".format(e, name))
                exception_occurred = True
        if name in main_packages_names:
            exists_action = args.exists_action
        else:
            exists_action = dep_exists_action
        try:
            imported_package = _handle_package_exists(
                mc, data, package, exists_action)
            if imported_package:
                imported_list.append(imported_package)
        except Exception as e:
            print("Error {0} occurred while installing package {1}".format(
                e, name))
            exception_occurred = True
    if imported_list:
        _print_package_list(imported_list)
    if exception_occurred:
        # NOTE(jose-phillips) Leave a Warning to users in case some packages
        # can be uploaded successfully.
        if imported_list:
            print("Warning: there were some errors during the operation.")
            sys.exit(1)
        else:
            sys.exit(1)


@utils.arg("id", metavar="<ID>",
           help="Package ID to update.")
@utils.arg('--is-public', type=_bool_from_str_strict, metavar='{true|false}',
           help='Make package available to users from other tenants.')
@utils.arg('--enabled', type=_bool_from_str_strict, metavar='{true|false}',
           help='Make package active and available for deployments.')
@utils.arg('--name', default=None, help='New name for the package.')
@utils.arg('--description', default=None, help='New package description.')
@utils.arg('--tags', metavar='<TAG>', nargs='*',
           default=None,
           help='A list of keywords connected to the application.')
def do_package_update(mc, args):
    """Update an existing package."""
    data = {}
    parameters = ('is_public', 'enabled',
                  'name', 'description',
                  'tags')
    for parameter in parameters:
        param_value = getattr(args, parameter, None)
        if param_value is not None:
            data[parameter] = param_value

    mc.packages.update(args.id, data)
    do_package_show(mc, args)


@utils.arg('filename', metavar='<FILE>',
           nargs='+',
           help='Bundle URL, bundle name, or path to the bundle file.')
@utils.arg('--is-public', action='store_true', default=False,
           help='Make packages available to users from other tenants.')
@utils.arg('--exists-action', default='', choices=['a', 's', 'u'],
           help='Default action when a package already exists.')
def do_bundle_import(mc, args):
    """Import a bundle.

    `FILE` can be either a path to a zip file, URL, or name from repo.
    If `FILE` is a local file, treat names of packages in a bundle as
    file names, relative to location of the bundle file. Requirements
    are first searched in the same directory.
    """
    total_reqs = collections.OrderedDict()
    for filename in args.filename:
        local_path = None
        if os.path.isfile(filename):
            _file = filename
            local_path = os.path.dirname(os.path.abspath(filename))
        else:
            print("Bundle file '{0}' does not exist, attempting to download"
                  "".format(filename))
            _file = utils.to_url(
                filename,
                base_url=args.murano_repo_url,
                path='bundles/',
                extension='.bundle',
            )

        try:
            bundle_file = utils.Bundle.from_file(_file)
        except Exception as e:
            print("Failed to create bundle for '{0}', reason: {1}".format(
                filename, e))
            continue

        data = {"is_public": args.is_public}

        try:
            for package in bundle_file.packages(
                    base_url=args.murano_repo_url, path=local_path):

                requirements = package.requirements(
                    base_url=args.murano_repo_url,
                    path=local_path,
                )
                total_reqs.update(requirements)
        except Exception:
            print("Can't parse bundle contents")
            continue

    imported_list = []

    for name, dep_package in total_reqs.items():
        image_specs = dep_package.images()
        if image_specs:
            print("Inspecting required images")
            try:
                imgs = utils.ensure_images(
                    glance_client=mc.glance_client,
                    image_specs=image_specs,
                    base_url=args.murano_repo_url,
                    local_path=local_path,
                    is_package_public=args.is_public)
                for img in imgs:
                    print("Added {0}, {1} image".format(
                        img['name'], img['id']))
            except Exception as e:
                print("Error {0} occurred while installing "
                      "images for {1}".format(e, name))
        try:
            imported_package = _handle_package_exists(
                mc, data, dep_package, args.exists_action)
            if imported_package:
                imported_list.append(imported_package)
        except exceptions.CommandError:
            raise
        except Exception as e:
            print("Error {0} occurred while "
                  "installing package {1}".format(e, name))
    if imported_list:
        _print_package_list(imported_list)


def _handle_save_packages(packages, dst, base_url, no_images):
    downloaded_images = []

    for name, pkg in packages.items():
        if not no_images:
            image_specs = pkg.images()
            for image_spec in image_specs:
                if not image_spec["Name"]:
                    print("Invalid image.lst file for {0} package. "
                          "'Name' section is absent.".format(name))
                    continue
                if image_spec["Name"] not in downloaded_images:
                    print("Package {0} depends on image {1}. "
                          "Downloading...".format(name, image_spec["Name"]))
                    try:
                        utils.save_image_local(image_spec, base_url, dst)
                        downloaded_images.append(image_spec["Name"])
                    except Exception as e:
                        print("Error {0} occurred while saving image {1}".
                              format(e, image_spec["Name"]))

        try:
            pkg.save(dst)
            print("Package {0} has been successfully saved".format(name))
        except Exception as e:
            print("Error {0} occurred while saving package {1}".format(
                e, name))


@utils.arg('filename', metavar='<BUNDLE>',
           help='Bundle URL, bundle name, or path to the bundle file.')
@utils.arg('-p', '--path', metavar='<PATH>',
           help='Path to the directory to store packages. If not set will use '
                'current directory.')
@utils.arg('--no-images', action='store_true', default=False,
           help='If set will skip images downloading.')
def do_bundle_save(mc, args):
    """Save a bundle.

    This will download a bundle of packages with all dependencies
    to specified path. If path doesn't exist it will be created.
    """

    bundle = args.filename
    base_url = args.murano_repo_url

    if args.path:
        if not os.path.exists(args.path):
            os.makedirs(args.path)
        dst = args.path
    else:
        dst = os.getcwd()

    total_reqs = collections.OrderedDict()

    if os.path.isfile(bundle):
        _file = bundle
    else:
        print("Bundle file '{0}' does not exist, attempting to download"
              .format(bundle))
        _file = utils.to_url(
            bundle,
            base_url=base_url,
            path='bundles/',
            extension='.bundle',
        )
    try:
        bundle_file = utils.Bundle.from_file(_file)
    except Exception as e:
        msg = "Failed to create bundle for {0}, reason: {1}".format(bundle, e)
        raise exceptions.CommandError(msg)

    for package in bundle_file.packages(base_url=base_url):
        requirements = package.requirements(base_url=base_url)
        total_reqs.update(requirements)

    no_images = getattr(args, 'no_images', False)

    _handle_save_packages(total_reqs, dst, base_url, no_images)

    try:
        bundle_file.save(dst, binary=False)
        print("Bundle file {0} has been successfully saved".format(bundle))
    except Exception as e:
        print("Error {0} occurred while saving bundle {1}".format(e, bundle))


@utils.arg('package', metavar='<PACKAGE>',
           nargs='+',
           help='Package URL or name.')
@utils.arg('-p', '--path', metavar='<PATH>',
           help='Path to the directory to store package. If not set will use '
                'current directory.')
@utils.arg('--package-version', default='',
           help='Version of the package to use from repository '
                '(ignored when saving with multiple packages).')
@utils.arg('--no-images', action='store_true', default=False,
           help='If set will skip images downloading.')
def do_package_save(mc, args):
    """Save a package.

    This will download package(s) with all dependencies
    to specified path. If path doesn't exist it will be created.
    """
    base_url = args.murano_repo_url

    if args.path:
        if not os.path.exists(args.path):
            os.makedirs(args.path)
        dst = args.path
    else:
        dst = os.getcwd()

    version = args.package_version
    if version and len(args.filename) >= 2:
        print("Requested to save more than one package, "
              "ignoring version.")
        version = ''

    total_reqs = collections.OrderedDict()
    for package in args.package:
        _file = utils.to_url(
            package,
            version=version,
            base_url=base_url,
            extension='.zip',
            path='apps/',
        )
        try:
            pkg = utils.Package.from_file(_file)
        except Exception as e:
            print("Failed to create package for '{0}', reason: {1}".format(
                package, e))
            continue
        total_reqs.update(pkg.requirements(base_url=base_url))

    no_images = getattr(args, 'no_images', False)

    _handle_save_packages(total_reqs, dst, base_url, no_images)


@utils.arg('id', metavar='<ID>',
           help='Environment ID to show applications from.')
@utils.arg('-p', '--path', metavar='<PATH>',
           help='Level of detalization to show. '
                'Leave empty to browse all applications in the environment.',
           default='/')
def do_app_show(mc, args):
    """List applications, added to specified environment."""
    if args.path == '/':
        apps = mc.services.list(args.id)
        formatters = {'id': lambda x: getattr(x, '?')['id'],
                      'type': lambda x: getattr(x, '?')['type']}
        field_labels = ['Id', 'Name', 'Type']
        fields = ['id', 'name', 'type']
        utils.print_list(apps, fields, field_labels, formatters=formatters)
    else:
        if not args.path.startswith('/'):
            args.path = '/' + args.path
        app = mc.services.get(args.id, args.path)

        # If app with specified path is not found, it is empty.
        if hasattr(app, '?'):
            formatters = {}
            for key in app.to_dict().keys():
                formatters[key] = utils.json_formatter
            utils.print_dict(app.to_dict(), formatters)
        else:
            raise exceptions.CommandError("Could not find application at path"
                                          " %s" % args.path)


@utils.arg('-t', '--template', metavar='<HEAT_TEMPLATE>',
           help='Path to the Heat template to import as '
                'an Application Definition.')
@utils.arg('-c', '--classes-dir', metavar='<CLASSES_DIRECTORY>',
           help='Path to the directory containing application classes.')
@utils.arg('-r', '--resources-dir', metavar='<RESOURCES_DIRECTORY>',
           help='Path to the directory containing application resources.')
@utils.arg('-n', '--name', metavar='<DISPLAY_NAME>',
           help='Display name of the Application in Catalog.')
@utils.arg('-f', '--full-name', metavar='<full-name>',
           help='Fully-qualified name of the Application in Catalog.')
@utils.arg('-a', '--author', metavar='<AUTHOR>', help='Name of the publisher.')
@utils.arg('--tags', help='A list of keywords connected to the application.',
           metavar='<TAG1 TAG2>', nargs='*')
@utils.arg('-d', '--description', metavar='<DESCRIPTION>',
           help='Detailed description for the Application in Catalog.')
@utils.arg('-o', '--output', metavar='<PACKAGE_NAME>',
           help='The name of the output file archive to save locally.')
@utils.arg('-u', '--ui', metavar='<UI_DEFINITION>',
           help='Dynamic UI form definition.')
@utils.arg('--type',
           help='Package type. Possible values: Application or Library.')
@utils.arg('-l', '--logo', metavar='<LOGO>', help='Path to the package logo.')
def do_package_create(mc, args):
    """Create an application package."""
    if args.template and args.classes_dir:
        raise exceptions.CommandError(
            "Provide --template for a HOT-based package, OR"
            " --classes-dir for a MuranoPL-based package")
    if not args.template and not args.classes_dir:
        raise exceptions.CommandError(
            "Provide --template for a HOT-based package, OR at least"
            " --classes-dir for a MuranoPL-based package")
    directory_path = None
    try:
        archive_name = args.output if args.output else None
        if args.template:
            directory_path = hot_package.prepare_package(args)
            if not archive_name:
                archive_name = os.path.basename(args.template)
                archive_name = os.path.splitext(archive_name)[0] + ".zip"
        else:
            directory_path = mpl_package.prepare_package(args)
            if not archive_name:
                archive_name = tempfile.mkstemp(
                    prefix="murano_", dir=os.getcwd())[1] + ".zip"

        _make_archive(archive_name, directory_path)
        print("Application package is available at " +
              os.path.abspath(archive_name))
    finally:
        if directory_path:
            shutil.rmtree(directory_path)


def _make_archive(archive_name, path):
    zip_file = zipfile.ZipFile(archive_name, 'w')
    for root, dirs, files in os.walk(path):
        for f in files:
            zip_file.write(os.path.join(root, f),
                           arcname=os.path.join(os.path.relpath(root, path),
                                                f))


def do_category_list(mc, args=None):
    """List all available categories."""
    if args is None:
        args = {}
    categories = mc.categories.list()
    _print_category_list(categories)


def _print_category_list(categories):
    field_labels = ["ID", "Name"]
    fields = ["id", "name"]
    utils.print_list(categories, fields, field_labels)


@utils.arg("id", metavar="<ID>",
           help="ID of a category(s) to show.")
def do_category_show(mc, args):
    """Display category details."""
    try:
        category = mc.categories.get(args.id)
        packages = mc.packages.filter(category=category.name)
        to_display = dict(id=category.id,
                          name=category.name,
                          packages=', '.join(p.name for p in packages))
        formatters = {'packages': utils.text_wrap_formatter}
        utils.print_dict(to_display, formatters)
    except common_exceptions.HTTPNotFound:
        print("Category id '{0}' not found". format(args.id))


@utils.arg("name", metavar="<CATEGORY_NAME>",
           help="Category name.")
def do_category_create(mc, args):
    """Create a category."""
    category = mc.categories.add({"name": args.name})
    _print_category_list([category])


@utils.arg("id", metavar="<ID>",
           nargs="+", help="ID of a category(ies) to delete.")
def do_category_delete(mc, args):
    """Delete a category."""
    failure_count = 0
    for category_id in args.id:
        try:
            mc.categories.delete(category_id)
        except common_exceptions.HTTPNotFound:
            failure_count += 1
            print("Failed to delete '{0}'; category not found".
                  format(category_id))
    if failure_count == len(args.id):
        raise exceptions.CommandError("Unable to find and delete any of the "
                                      "specified categories.")
    do_category_list(mc)


@utils.arg("class_name", metavar="<CLASS>", help="Class FQN")
@utils.arg("method_names", metavar="<METHOD>", help="Method name", nargs='*')
@utils.arg("--package-name", default=None,
           help="FQN of the package where the class is located")
@utils.arg("--class-version", default='=0',
           help="Class version or version range (version spec)")
def do_class_schema(mc, args):
    """Display class schema"""
    schema = mc.schemas.get(args.class_name, args.method_names,
                            class_version=args.class_version,
                            package_name=args.package_name)
    print(utils.json_formatter(schema.data))



# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from glanceclient.common import http
from glanceclient.common import utils

from muranoclient.glance import artifacts


class Client(object):
    """Client for the OpenStack glance-glare API.

    :param string endpoint: A user-supplied endpoint URL for the glance
                            service.
    :param string token: Token for authentication.
    :param integer timeout: Allows customization of the timeout for client
                            http requests. (optional)
    """

    def __init__(self, endpoint, type_name, type_version, **kwargs):
        endpoint, version = utils.strip_version(endpoint)
        self.version = version or '0.1'
        self.http_client = http.HTTPClient(endpoint, **kwargs)

        self.type_name = type_name
        self.type_version = type_version

        self.artifacts = artifacts.Controller(self.http_client,
                                              self.type_name,
                                              self.type_version,
                                              self.version)



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import urllib

from muranoclient.common import base


class Category(base.Resource):
    def __repr__(self):
        return "<Category %s>" % self._info

    def data(self, **kwargs):
        return self.manager.data(self, **kwargs)


class CategoryManager(base.Manager):
    resource_class = Category

    def list(self, **kwargs):
        """Get category list with pagination support.

        :param sort_keys: an array of fields used to sort the list (string)
        :param sort_dir: 'asc' or 'desc' for ascending or descending sort
        :param limit: maximum number of categories to return
        :param marker: begin returning categories that appear later in the
                       category list than that represented by this marker id
        """

        params = {}
        for key, value in kwargs.items():
            if value:
                params[key] = value

        url = '/v1/catalog/categories?{0}'.format(
            urllib.parse.urlencode(params, True))
        return self._list(url, response_key='categories')

    def get(self, id):
        return self._get('/v1/catalog/categories/{0}'.format(id))

    def add(self, data):
        return self._create('/v1/catalog/categories', data)

    def delete(self, id):
        return self._delete('/v1/catalog/categories/{0}'.format(id))
